from pathlib import Path
import torch
import random
import yaml
import numpy
from torch_geometric.loader import DataLoader
from .datasets import ActsTrack, PLBind, Tau3Mu, SynMol
from torch_geometric.nn import knn_graph, radius_graph
from torch.utils.data import Subset

DATASET_CANDIS = ['tau3mu', 'plbind', 'synmol', 'actstrack']

def syn_transform(data):
    edge_index = knn_graph(data.pos, k=5, batch=data.batch, loop=True)
    data.edge_index = edge_index
    return data

def act_transform(data):
    # pos = data.pos / 2955.5000 * 100
    norm_pos = data.pos.norm(dim=-1, keepdim=True)
    pos = data.pos / norm_pos.clamp(min=1e-6)
    edge_index = knn_graph(pos, k=5, batch=data.batch, loop=True)
    data.edge_index = edge_index
    return data

def tau_transform(data):
    edge_index = radius_graph(data.pos, r=1.0, batch=data.batch, loop=True)
    data.edge_index = edge_index
    return data

def plb_transform(data):
    edge_index = radius_graph(data.pos, r=2.0, batch=data.batch, loop=True)
    data.edge_index = edge_index
    return data


class ScienceDataset:
    _initialized_datasets = {}

    def __init__(self, *args, **kwargs):
        raise EnvironmentError(
            f"{self.__class__.__name__} is designed to be instantiated "
            f"using the `{self.__class__.__name__}.from_name(method_name)`."
        )

    @classmethod
    def filter_signal_class(cls, dataset, target_label=torch.tensor([[1]])):
        target_label = target_label.to(dtype=torch.float32)
        indices = [i for i, data in enumerate(dataset) if torch.equal(data.y, target_label)]
        return Subset(dataset, indices)

    @classmethod
    def from_name(cls, name, config=None) -> None:
        if name in cls._initialized_datasets:
            return cls._initialized_datasets[name]

        assert name in DATASET_CANDIS, f"Dataset name {name} is not supported, choose from {DATASET_CANDIS}"
        if config is None:
            cfg_path = Path(__file__).parent / 'configs' / f'{name}.yml'
            config = yaml.safe_load((cfg_path).open('r'))['data']
        data_dir = Path(config['data_dir'])
        if 'actstrack' in name:
            tesla = '2T' if len(name.split('_')) == 1 else name.split('_')[-1]
            dataset = ActsTrack(data_dir / 'actstrack', tesla=tesla, data_config=config, transform=act_transform)
        elif name == 'tau3mu':
            dataset = Tau3Mu(data_dir / 'tau3mu', data_config=config, transform=tau_transform)
        elif name == 'synmol':
            dataset = SynMol(data_dir / 'synmol', data_config=config, transform=syn_transform)
        elif name == 'plbind':
            dataset = PLBind(data_dir / 'plbind', data_config=config, transform=plb_transform, n_jobs=32, debug=False)

        cls._initialized_datasets[name] = dataset
        return dataset

    def get_train_loader(self, batch_size):
        follow_batch = 'x_lig' if self.dataset_name == 'plbind' else None
        train_ds = [self[self.index_split['train']]]
        train_loader = DataLoader(train_ds, batch_size=batch_size, shuffle=True, follow_batch=follow_batch)
        return train_loader

def get_data_loaders(dataset_name, batch_size, data_config, dataset_seed, num_workers, device):
    data_dir = Path(data_config['data_dir'])
    assert dataset_name in ['tau3mu', 'plbind', 'synmol'] or 'acts' in dataset_name

    if 'actstrack' in dataset_name:
        def act_transform(data):
            # pos = data.pos / 2955.5000 * 100
            norm_pos = data.pos.norm(dim=-1, keepdim=True)
            pos = data.pos / norm_pos.clamp(min=1e-6)
            edge_index = knn_graph(pos, k=5, batch=data.batch, loop=True)
            data.edge_index = edge_index
            return data
        tesla = '2T' if len(dataset_name.split('_')) == 1 else dataset_name.split('_')[-1]
        dataset = ActsTrack(data_dir / 'actstrack', tesla=tesla, data_config=data_config, seed=dataset_seed, transform=act_transform, device=device)
        loaders, test_set = get_loaders(batch_size, dataset=dataset, idx_split=dataset.idx_split, num_workers=num_workers)

    elif dataset_name == 'tau3mu':
        def tau_transform(data):
            edge_index = radius_graph(data.pos, r=1.0, batch=data.batch, loop=True)
            data.edge_index = edge_index
            return data
        dataset = Tau3Mu(data_dir / 'tau3mu', data_config=data_config, seed=dataset_seed, transform=tau_transform, device=device)
        loaders, test_set = get_loaders(batch_size, dataset=dataset, idx_split=dataset.idx_split, num_workers=num_workers)

    elif dataset_name == 'synmol':
        def syn_transform(data):
            edge_index = knn_graph(data.pos, k=5, batch=data.batch, loop=True)
            data.edge_index = edge_index
            return data
        dataset = SynMol(data_dir / 'synmol', data_config=data_config, seed=dataset_seed, transform=syn_transform, device=device)
        loaders, test_set = get_loaders(batch_size, dataset=dataset, idx_split=dataset.idx_split, num_workers=num_workers)

    elif dataset_name == 'plbind':
        dataset = PLBind(data_dir / 'plbind', data_config=data_config, device=device, n_jobs=32, debug=False)
        loaders, test_set = get_loaders(batch_size, dataset=dataset, idx_split=dataset.idx_split, dataset_name=dataset_name, num_workers=num_workers)

    return loaders, test_set, dataset

def seed_worker(worker_id):
    worker_seed = torch.initial_seed() % 2 ** 32
    numpy.random.seed(worker_seed)
    random.seed(worker_seed)

def get_loaders(batch_size, dataset, num_workers=8):
    idx_split = dataset.idx_split
    dataset_name = dataset.dataset_name
    #
    # def collate_gpu(batch, device='cpu'):
    #     x, t = torch.utils.data.dataloader.default_collate(batch)
    #     return x.to(device=device), t.to(device=device)
    # from functools import partial
    # collate_certain_gpu = partial(collate_gpu, device=device)

    follow_batch = None if dataset_name != 'plbind' else ['x_lig']
    # train_index = idx_split["train"]#[1335:1400]

    train_loader = DataLoader(dataset[idx_split["train"]], batch_size=batch_size, shuffle=True, num_workers=num_workers, worker_init_fn=seed_worker, follow_batch=follow_batch)
    valid_loader = DataLoader(dataset[idx_split["valid"]], batch_size=batch_size, shuffle=False, num_workers=num_workers, worker_init_fn=seed_worker, follow_batch=follow_batch)
    test_loader = DataLoader(dataset[idx_split["test"]], batch_size=batch_size, shuffle=False, num_workers=num_workers, worker_init_fn=seed_worker, follow_batch=follow_batch)

    # test_set = dataset.copy(idx_split["test"])  # For visualization
    return {'train': train_loader, 'valid': valid_loader, 'test': test_loader}



import torch
import torch.nn as nn
from torch_scatter import scatter
import numpy as np


class BaseRandom(nn.Module):

    def __init__(self):
        super().__init__()
        # self.clf = clf
        # self.extractor = extractor
        # self.criterion = criterion
        # self.device = next(self.parameters()).device

    def warming(self, data):
        clf_logits = self.clf(data)
        pred_loss = self.criterion(clf_logits, data.y.float())
        return pred_loss, {'loss': pred_loss.item(), 'pred': pred_loss.item()}, clf_logits, None
        
    def train(self, dataset):
        pass
    
    def predict(self, data):
        original_clf_logits = self.clf(data)
        return original_clf_logits

    def forward_pass(self, data, epoch, do_sampling):
        raise NotImplementedError("Please use the implementation of specific interpretation method")
        node_labels = data.node_label
        emb, edge_index = self.clf.get_emb(data)

        # node_attn_log_logits = self.extractor(emb, batch=data.batch, pool_out_lig=None)

        node_attn = torch.rand_like(edge_index[0])
            # self.sampling(node_attn_log_logits, do_sampling)
        # edge_attn = self.node_attn_to_edge_attn(node_attn, edge_index)
        # masked_clf_logits = self.clf(data, edge_attn=edge_attn, node_attn=node_attn)
        original_clf_logits = self.clf(data)

        # loss, loss_dict = self.__loss__(node_attn_log_logits.sigmoid(), masked_clf_logits, data.y, epoch)
        return -1, {'pred': -1}, original_clf_logits, original_clf_logits, node_attn.reshape(-1)

    @staticmethod
    def node_attn_to_edge_attn(node_attn, edge_index):
        src_attn = node_attn[edge_index[0]]
        dst_attn = node_attn[edge_index[1]]
        edge_attn = src_attn * dst_attn
        return edge_attn

    @staticmethod
    def min_max_scalar(attn, const=0.5):
        if torch.isnan(attn).any():
            print(f"Remind there are nan in {torch.isnan(attn).nonzero()}!")
            attn = torch.where(torch.isnan(attn), torch.full_like(attn, 0), attn)

        if attn.max() - attn.min() == 0:
            return torch.full_like(attn, fill_value=const)
        else:
            return (attn - attn.min()) / (attn.max() - attn.min())


class LabelPerturb(nn.Module):

    def __init__(self, model, mode):
        super().__init__()
        self.name = 'label_perturb'
        self.mode = mode
        self.clf = model
        self.device = next(self.parameters()).device

    def forward_pass(self, data, epoch, do_sampling):
        node_labels = data.node_label
        node_attn = torch.rand_like(node_labels)
        # emb, edge_index = self.clf.get_emb(data)
        original_clf_logits = self.clf(data)

        if self.mode == 1:
            node_attn[torch.where(node_labels == 1)[0]] = 1
        else:
            assert self.mode == 0
            node_attn[torch.where(node_labels == 1)[0]] = 0

        # from sklearn.metrics import roc_auc_score
        # msk_emb = emb.clone()
        # msk_emb[torch.where(data.node_label == 0)] = torch.zeros((msk_emb.shape[1],))
        # mask_clf_logits = self.clf.get_pred_from_emb(msk_emb, data.batch)
        # org_pred = original_clf_logits.sigmoid()
        # msk_pred = mask_clf_logits.sigmoid()
        # org_score = roc_auc_score(data.y.cpu(), original_clf_logits.sigmoid().cpu())
        # msk_score = roc_auc_score(data.y.cpu(), mask_clf_logits.sigmoid().cpu())
        # roc_auc_score(data.y.cpu(), original_clf_logits.sigmoid().cpu())

        return -1, {'pred': -1}, original_clf_logits, node_attn.reshape(-1)



import torch
import torch.nn as nn
from torch.nn import functional as F
from torch_geometric.nn import knn_graph, radius_graph
from torch_geometric.nn import global_mean_pool, global_add_pool, global_max_pool
from .baselines import *
from .backbones import DGCNN, PointTransformer, EGNN
from .utils import ExtractorMLP, MLP, CoorsNorm, get_optimizer, log_epoch
from pathlib import Path
import yaml
from .utils import inherent_models, post_hoc_explainers, post_hoc_attribution
from .get_data import get_loaders, ScienceDataset
from .trainer import run_one_epoch, train_one_batch, select_augmentation
from tqdm import tqdm

POST_ATTRIBUTOR_MAPPING = {
    'gradcam': GradCAM,
    'gradx': GradX,
    'gnnlrp': GNNLRP,
    'ig': InterGrad,
    'subgraphx': SubgraphX,
    'pgmexplainer': PGMExplainer,
}

POSTHOC_PARAMETRIZED_MODEL_MAPPING = {
    'gnnexplainer': GNNExplainer,
    'pgexplainer': PGExplainer,
}

INHERENT_MODEL_MAPPING = {
    'lri_bern': LRIBern,
    'lri_gaussian': LRIGaussian,
    'asap': ASAP,
    'vgib': VGIB,
    'ciga': CIGA,
}

class _XBaseModel:
    # Base class for auto models.
    _model_mapping = None

    def __init__(self, config, **kwargs):
        self.baseline = self.from_config(config)
        self.config = config
        # raise EnvironmentError(
        #     f"{self.__class__.__name__} is designed to be instantiated "
        #     f"using the `{self.__class__.__name__}.from_name(method_name)` or "
        #     f"`{self.__class__.__name__}.from_config(config)` methods."
        # )

    def __repr__(self):
        return f'{self.baseline}'

    @classmethod
    def _prepare_clf_extractor(cls, dataset, config):
        model_name = config['model']
        method_name = config['method']
        method_config = config['hyperparameter']
        cfg_path = Path(__file__).parent / 'configs' / f'{dataset.dataset_name}.yml'
        model_config = yaml.safe_load((cfg_path).open('r'))['model'][f'{model_name}']
        
        clf = Model(model_name, model_config, method_name, method_config, dataset)
        if method_name == 'asap':
            extractor = ASAPooling(model_config['hidden_size'], ratio=method_config['casual_ratio'], dropout=model_config['dropout_p'])
        elif method_name in inherent_models + ['pgexplainer']:
            extractor =  ExtractorMLP(model_config['hidden_size'], model_config, getattr(dataset, "use_lig_info", False))
        else:
            extractor = nn.Identity()

        return clf, extractor

    @classmethod
    def from_config(cls, config):
        dataset_info = ScienceDataset.from_name(config['dataset'])
        criterion = F.binary_cross_entropy_with_logits
        clf, extractor = cls._prepare_clf_extractor(dataset_info, config)

        name = config['method']
        method_config = config['hyperparameter']
        method_config['signal_class'] = dataset_info.signal_class

        model_class = cls._model_mapping[name]
        if 'grad' in name or name == 'gnnlrp':
            model = model_class(clf, criterion, method_config)
        else:
            model = model_class(clf, extractor, criterion, method_config)
        return model
    
    def warmup(self, dataset):
        trn_cfg = self.config['training']
        dataloader = get_loaders(trn_cfg['batch_size'], dataset)
        data_aug = select_augmentation(dataset.dataset_name)
        train_loader = dataloader['train']
        valid_loader = dataloader['valid']

        optimizer = get_optimizer(self.baseline.clf, self.baseline.extractor, trn_cfg, self.config['method'], warmup=True)
        loader_len = len(train_loader)
        pbar = tqdm(train_loader)
        phase = 'warm'
        for epoch in range(trn_cfg['warmup']):
            for idx, data in enumerate(pbar):
                data = data_aug(data, "train", train_loader, idx, loader_len=loader_len)
                loss_dict, clf_logits, attn = train_one_batch(self.baseline, optimizer, data, epoch, phase)
                desc = log_epoch(epoch, phase, loss_dict)
                pbar.set_description(desc) 


    def explain(self, data):
        self.baseline.clf.eval()
        self.baseline.extractor.eval() if hasattr(self.baseline, 'extractor') else None

        self.baseline.start_tracking() if 'grad' in self.config['method'] or self.config['method'] == 'gnnlrp' else None

        do_sampling = True if self.config['method'] == 'pgexplainer' else False # we find this is better for BernMaskP
        loss, loss_dict, infer_clf_logits, node_attn = self.baseline.forward_pass(data, epoch=0, do_sampling=do_sampling)
        data.node_imp = node_attn
        return data

    def predict(self, sample):
        return self.baseline(sample)

class PosthocAttributor(_XBaseModel):
    _model_mapping = POST_ATTRIBUTOR_MAPPING
    def train(self, dataset):
        trn_cfg = self.config['training']
        dataloader = get_loaders(trn_cfg['batch_size'], dataset)
        data_aug = select_augmentation(dataset.dataset_name)
        train_loader = dataloader['train']
        valid_loader = dataloader['valid']
        loader_len = len(train_loader)

        print('Begin to train base model for prediction.')

        pbar = tqdm(train_loader)
        optimizer = get_optimizer(self.baseline.clf, None, trn_cfg, self.config['method'], warmup=True)
        phase = 'warm'
        for epoch in range(trn_cfg['epoch']):
            for idx, data in enumerate(pbar):
                data = data_aug(data, "train", train_loader, idx, loader_len=loader_len)
                loss_dict, clf_logits, attn = train_one_batch(self.baseline, optimizer, data, epoch, phase)
                desc = log_epoch(epoch, phase, loss_dict)
                pbar.set_description(desc) 

        print(f"Method {self.config['method']} do not need to train additional module. Training is completed.")


class PosthocParametrizedModel(_XBaseModel):
    _model_mapping = POSTHOC_PARAMETRIZED_MODEL_MAPPING
    def train(self, dataset):
        trn_cfg = self.config['training']
        dataloader = get_loaders(trn_cfg['batch_size'], dataset)
        data_aug = select_augmentation(dataset.dataset_name)
        train_loader = dataloader['train']
        valid_loader = dataloader['valid']
        loader_len = len(train_loader)

        print('Begin to train base model for prediction.')

        pbar = tqdm(train_loader)
        optimizer = get_optimizer(self.baseline.clf, None, trn_cfg, self.config['method'], warmup=True)
        phase = 'warm'
        for epoch in range(trn_cfg['epoch']):
            for idx, data in enumerate(pbar):
                data = data_aug(data, "train", train_loader, idx, loader_len=loader_len)
                loss_dict, clf_logits, attn = train_one_batch(self.baseline, optimizer, data, epoch, phase)
                desc = log_epoch(epoch, phase, loss_dict)
                pbar.set_description(desc) 

        print("Continue to train additional module for explanation.")
        
        pbar = tqdm(train_loader)
        optimizer = get_optimizer(self.baseline.clf, self.baseline.extractor, trn_cfg, self.config['method'], warmup=False)
        phase = 'train' 
        for epoch in range(trn_cfg['epoch']):
            for idx, data in enumerate(pbar):
                data = data_aug(data, "train", train_loader, idx, loader_len=loader_len)
                loss_dict, clf_logits, attn = train_one_batch(self.baseline, optimizer, data, epoch, phase)
                desc = log_epoch(epoch, phase, loss_dict)
                pbar.set_description(desc) 


class PosthocMethod(_XBaseModel):
    # _model_mapping = POSTHOC_PARAMETRIZED_MODEL_MAPPING | POST_ATTRIBUTOR_MAPPING
    _model_mapping = {**POSTHOC_PARAMETRIZED_MODEL_MAPPING, **POST_ATTRIBUTOR_MAPPING}  
    def train(self, dataset):
        trn_cfg = self.config['training']
        dataloader = get_loaders(trn_cfg['batch_size'], dataset)
        data_aug = select_augmentation(dataset.dataset_name)
        train_loader = dataloader['train']
        valid_loader = dataloader['valid']
        loader_len = len(train_loader)

        print('Begin to train base model for prediction.')

        pbar = tqdm(train_loader)
        optimizer = get_optimizer(self.baseline.clf, None, trn_cfg, self.config['method'], warmup=True)
        phase = 'warm'
        for epoch in range(trn_cfg['epoch']):
            for idx, data in enumerate(pbar):
                data = data_aug(data, "train", train_loader, idx, loader_len=loader_len)
                loss_dict, clf_logits, attn = train_one_batch(self.baseline, optimizer, data, epoch, phase)
                desc = log_epoch(epoch, phase, loss_dict)
                pbar.set_description(desc) 


        if self.config['method'] in POST_ATTRIBUTOR_MAPPING.keys():
            print(f"Method {self.config['method']} do not need to train additional module. Training is completed.")
            return
        else:
            print("Continue to train additional module for explanation.")

        pbar = tqdm(train_loader)
        optimizer = get_optimizer(self.baseline.clf, self.baseline.extractor, trn_cfg, self.config['method'], warmup=False)
        phase = 'train' 
        for epoch in range(trn_cfg['epoch']):
            for idx, data in enumerate(pbar):
                data = data_aug(data, "train", train_loader, idx, loader_len=loader_len)
                loss_dict, clf_logits, attn = train_one_batch(self.baseline, optimizer, data, epoch, phase)
                desc = log_epoch(epoch, phase, loss_dict)
                pbar.set_description(desc) 


class InherentModel(_XBaseModel):
    _model_mapping = INHERENT_MODEL_MAPPING
    def train(self, dataset):
        trn_cfg = self.config['training']
        dataloader = get_loaders(trn_cfg['batch_size'], dataset)
        data_aug = select_augmentation(dataset.dataset_name)
        train_loader = dataloader['train']
        valid_loader = dataloader['valid']
        optimizer = get_optimizer(self.baseline.clf, self.baseline.extractor, trn_cfg, self.config['method'], warmup=False)
        loader_len = len(train_loader)
        pbar = tqdm(train_loader)
        phase = 'train'
        for epoch in range(trn_cfg['epoch']):
            for idx, data in enumerate(pbar):
                data = data_aug(data, "train", train_loader, idx, loader_len=loader_len)
                loss_dict, clf_logits, attn = train_one_batch(self.baseline, optimizer, data, epoch, phase)
                desc = log_epoch(epoch, phase, loss_dict)
                pbar.set_description(desc) 


class Model(nn.Module):
    def __init__(self, model_name, model_config, method_name, method_config, dataset):
        super().__init__()
        
        assert dataset.dataset_name in ['tau3mu', 'plbind', 'synmol'] or 'actstrack' in dataset.dataset_name
        # self.dataset = dataset
        self.dataset_name = dataset.dataset_name
        self.method_name = method_name
        self.one_encoder = method_config.get('one_encoder', True)
        self.covar_dim = method_config.get('covar_dim', None)
        self.pos_coef = method_config.get('pos_coef', None)
        self.kr = method_config.get('kr', None)
        if method_name == 'psat':
            assert self.pos_coef is not None and self.kr is not None

        out_dim = 1 if dataset.num_classes == 2 else dataset.num_classes
        hidden_size = model_config['hidden_size']
        dropout_p = model_config['dropout_p']
        norm_type = model_config['norm_type']
        act_type = model_config['act_type']

        if model_name == 'dgcnn':
            Model = DGCNN
        elif model_name == 'pointtrans':
            Model = PointTransformer
        elif model_name == 'egnn':
            Model = EGNN
        else:
            raise NotImplementedError

        if model_config['pool'] == 'mean':
            self.pool = global_mean_pool
        elif model_config['pool'] == 'max':
            self.pool = global_max_pool
        elif model_config['pool'] == 'add':
            self.pool = global_add_pool
        else:
            raise NotImplementedError

        dataset.feat_info['edge_categorical_feat'], dataset.feat_info['edge_scalar_feat'] = [], dataset.pos_dim+1
        if self.dataset_name == 'plbind':
            dataset.feat_info_lig['edge_categorical_feat'], dataset.feat_info_lig['edge_scalar_feat'] = [], dataset.pos_dim+1

        raw_pos_dim = dataset.pos_dim
        if dataset.feature_type == 'only_pos':
            dataset.x_dim = 0
            dataset.x_lig_dim = 0
        elif dataset.feature_type == 'only_x':
            dataset.pos_dim = 0
            dataset.pos_lig_dim = 0
        elif dataset.feature_type == 'only_ones':
            dataset.x_dim = 0
            dataset.x_lig_dim = 0
            dataset.pos_dim = 0
            dataset.pos_lig_dim = 0
        else:
            assert dataset.feature_type == 'both_x_pos'

        aux_info = {'raw_pos_dim': raw_pos_dim, 'dataset_name': dataset.dataset_name}
        self.coors_norm = CoorsNorm()
        self.mlp_out = MLP([hidden_size, hidden_size * 2, hidden_size, out_dim], dropout_p, norm_type, act_type)
        self.spu_mlp_out = MLP([hidden_size, hidden_size * 2, hidden_size, out_dim], dropout_p, norm_type, act_type) \
            if method_name == 'dir' else None

        if dataset.dataset_name != 'plbind':
            self.model = Model(dataset.x_dim, dataset.pos_dim, model_config, dataset.feat_info, aux_info=aux_info)
            self.emb_model = Model(dataset.x_dim, dataset.pos_dim, model_config, dataset.feat_info, aux_info=aux_info) if not self.one_encoder else None
        else:
            self.model = Model(dataset.x_dim, dataset.pos_dim, model_config, dataset.feat_info, dataset.n_categorical_feat_to_use, dataset.n_scalar_feat_to_use, aux_info=aux_info)
            self.emb_model = Model(dataset.x_dim, dataset.pos_dim, model_config, dataset.feat_info, aux_info=aux_info) if not self.one_encoder else None
            self.model_lig = Model(dataset.x_lig_dim, dataset.pos_lig_dim, model_config, dataset.feat_info_lig, dataset.n_categorical_feat_to_use_lig, dataset.n_scalar_feat_to_use_lig, aux_info=aux_info, is_lig=True)

        self.dim_mapping = nn.Linear(1, 8)
        self.message_weights = ExtractorMLP(8, model_config, False, out_dim=1)

    def forward(self, data, edge_attn=None, node_noise=None):
        x, pos, edge_index, edge_attr = self.calc_geo_feat(data, node_noise, self.method_name)
        # edge_attr = self.calc_edge_attr(data.pos, data.edge_index)
        edge_attn = self.get_message_weights(x, pos, edge_index, data.batch) if self.method_name == 'psat' else edge_attn
        if self.dataset_name != 'plbind':
            emb = self.model(x, pos, edge_attr, edge_index, data.batch, edge_attn=edge_attn)
            pool_out = self.pool(emb, batch=data.batch)
        else:
            _, _, edge_index_lig, edge_attr_lig = self.calc_geo_feat(data, None, self.method_name, is_lig=True)
            emb_rec = self.model(x, pos, edge_attr, edge_index, data.batch, edge_attn=edge_attn)
            emb_lig = self.model_lig(data.x_lig, data.pos_lig, edge_attr_lig, edge_index_lig, data.x_lig_batch)
            pool_out_rec, pool_out_lig = self.pool(emb_rec, batch=data.batch), self.pool(emb_lig, batch=data.x_lig_batch)
            pool_out = pool_out_rec + pool_out_lig
        return self.mlp_out(pool_out)

    def get_pred_from_emb(self, emb, batch):
        pool_out = self.pool(emb, batch=batch)
        return self.mlp_out(pool_out)

    def get_pred_from_spu_emb(self, emb, batch):
        pool_out = self.pool(emb, batch=batch)
        return self.spu_mlp_out(pool_out)

    def get_emb(self, data, edge_attn=None):
        x, pos, edge_index, edge_attr = self.calc_geo_feat(data, None, self.method_name)
        # edge_attr = self.calc_edge_attr(data.pos, data.edge_index)
        edge_attn = self.get_message_weights(x, pos, edge_index, data.batch) if self.method_name == 'psat' else None
        #! The former line has been changed: else edge_attn -> else None
        if self.one_encoder:
            emb = self.model(x, pos, edge_attr, edge_index, data.batch, edge_attn=edge_attn)
        else:
            emb = self.emb_model(x, pos, edge_attr, edge_index, data.batch, edge_attn=edge_attn)

        if self.dataset_name == 'plbind':
            _, _, edge_index_lig, edge_attr_lig = self.calc_geo_feat(data, None, self.method_name, is_lig=True)
            pool_out_lig = self.pool(self.model_lig(data.x_lig, data.pos_lig, edge_attr_lig, edge_index_lig, data.x_lig_batch), batch=data.x_lig_batch)  # pool_out_lig
            emb = emb + pool_out_lig[data.batch] # use_lig_info
        return emb, edge_index

    def get_message_weights(self, x, pos, edge_index, batch):
        col, row = edge_index
        dist = torch.norm(pos[col] - pos[row], dim=1, p=2, keepdim=True)
        input_feat = self.dim_mapping(dist)
        return self.message_weights(input_feat, batch[col]).sigmoid()

    def calc_geo_feat(self, data, node_noise, method_name, is_lig=False):
        x, pos, batch = (data.x_lig, data.pos_lig, data.x_lig_batch) if is_lig else (data.x, data.pos, data.batch)
        if 'actstrack' in self.dataset_name:
            pos = pos / 2955.5000 * 100 if self.pos_coef is None else pos / 2955.5000 * self.pos_coef
            pos = self.add_noise(pos, node_noise)

            pos = self.coors_norm(pos)
            edge_index = data.edge_index if 'lri' not in method_name else knn_graph(pos, k=5 if self.kr is None else int(self.kr), batch=batch, loop=True)
            edge_attr = self.calc_edge_attr(pos, edge_index)

        elif self.dataset_name == 'tau3mu':
            pos = pos * 1.0 if self.pos_coef is None else pos * self.pos_coef
            pos = self.add_noise(pos, node_noise)

            edge_index = data.edge_index if 'lri' not in method_name else radius_graph(pos, r=1.0 if self.kr is None else self.kr * self.pos_coef, loop=True, batch=batch)
            edge_attr = self.calc_edge_attr(pos, edge_index)

        elif self.dataset_name == 'synmol':
            pos = pos * 5.0 if self.pos_coef is None else pos * self.pos_coef
            pos = self.add_noise(pos, node_noise)

            edge_index = data.edge_index if 'lri' not in method_name else knn_graph(pos, k=5 if self.kr is None else int(self.kr), batch=batch, loop=True)
            edge_attr = self.calc_edge_attr(pos, edge_index)

        elif self.dataset_name == 'plbind':
            if is_lig:
                edge_index = radius_graph(pos, r=2.0, loop=True, batch=batch)
                edge_attr = self.calc_edge_attr(pos, edge_index)
            else:
                pos = pos * 1.0 if self.pos_coef is None else pos * self.pos_coef
                pos = self.add_noise(pos, node_noise)

                edge_index = data.edge_index if 'lri' not in method_name else knn_graph(pos, k=5 if self.kr is None else int(self.kr), flow='target_to_source', loop=True, batch=batch)
                edge_attr = self.calc_edge_attr(pos, edge_index)

        return x, pos, edge_index, edge_attr

    def add_noise(self, pos, node_noise):
        if node_noise is not None:
            pos[:, :self.covar_dim] = pos[:, :self.covar_dim] + node_noise
        return pos

    def calc_edge_attr(self, pos, edge_index):
        row, col = edge_index
        rel_dist = torch.norm(pos[row] - pos[col], dim=1, p=2, keepdim=True)
        coord_diff = pos[row] - pos[col]
        edge_dir = coord_diff / (rel_dist + 1e-6)
        edge_attr = torch.cat([rel_dist, edge_dir], dim=1)
        return edge_attr



import copy
import torch
import numpy as np
from typing import Callable, Union
from scipy.special import comb
from itertools import combinations
import torch.nn.functional as F
from torch_geometric.utils import to_networkx
from torch_geometric.data import Data, Batch, Dataset
from torch_geometric.loader import DataLoader
import networkx as nx
#from .utils.testing_datasets import MarginalSubgraphDataset

'''
Code adapted from Dive into Graphs (DIG)
Code: https://github.com/divelab/DIG
'''

empty_tuple = tuple()

class MarginalSubgraphDataset(Dataset):
    """ Collect pair-wise graph data to calculate marginal contribution. """
    def __init__(self, data, exclude_mask, include_mask, subgraph_build_func):
        super().__init__()
        self.num_nodes = data.num_nodes
        # self.X = data.x
        # self.edge_index = data.edge_index
        # self.pos = data.pos
        self.data = data
        self.device = data.x.device

        self.label = data.y
        self.exclude_mask = torch.tensor(exclude_mask).type(torch.float32).to(self.device)
        self.include_mask = torch.tensor(include_mask).type(torch.float32).to(self.device)
        self.subgraph_build_func = subgraph_build_func

    def len(self):
        return self.exclude_mask.shape[0]

    def get(self, idx):
        exclude_graph_X, exclude_pos, exclude_graph_edge_index = self.subgraph_build_func(self.data, self.exclude_mask[idx])
        include_graph_X, include_pos, include_graph_edge_index = self.subgraph_build_func(self.data, self.include_mask[idx])
        exclude_data = Data(x=exclude_graph_X, edge_index=exclude_graph_edge_index, pos=exclude_pos)
        include_data = Data(x=include_graph_X, edge_index=include_graph_edge_index, pos=include_pos)
        return exclude_data, include_data

def GnnNets_GC2value_func(gnnNets, target_class, forward_kwargs = {}):
    def value_func(batch):
        with torch.no_grad():
            #logits = gnnNets(data=batch)
            logits = gnnNets(batch.x, batch.edge_index, **forward_kwargs)
            print(forward_kwargs)
            print(batch.batch)
            probs = F.softmax(logits, dim=-1)
            print(probs, probs.shape)
            score = probs[:, target_class]
        return score
    return value_func


def GnnNets_NC2value_func(gnnNets_NC, node_idx: Union[int, torch.Tensor], target_class: torch.Tensor):
    def value_func(data):
        with torch.no_grad():
            #logits = gnnNets_NC(data=data)
            logits = gnnNets_NC(data.x, data.edge_index)
            probs = F.softmax(logits, dim=-1)
            # select the corresponding node prob through the node idx on all the sampling graphs
            batch_size = data.batch.max() + 1
            print(batch_size)
            probs = probs.reshape(batch_size, -1, probs.shape[-1])
            score = probs[:, node_idx, target_class]
            return score
    return value_func


def get_graph_build_func(build_method):
    if build_method.lower() == 'zero_filling':
        return graph_build_zero_filling
    elif build_method.lower() == 'split':
        return graph_build_split
    else:
        raise NotImplementedError


def marginal_contribution(data: Data, exclude_mask: np.ndarray, include_mask: np.ndarray,
                          value_func, subgraph_build_func):
    """ Calculate the marginal value for each pair. Here exclude_mask and include_mask are node mask. """
    marginal_subgraph_dataset = MarginalSubgraphDataset(data, exclude_mask, include_mask, subgraph_build_func)
    dataloader = DataLoader(marginal_subgraph_dataset, batch_size=256, shuffle=False, pin_memory=False, num_workers=0)

    marginal_contribution_list = []

    for exclude_data, include_data in dataloader:
        exclude_values = value_func(exclude_data)
        include_values = value_func(include_data)
        margin_values = include_values - exclude_values
        marginal_contribution_list.append(margin_values)

    marginal_contributions = torch.cat(marginal_contribution_list, dim=0)
    return marginal_contributions


def graph_build_zero_filling(data, node_mask: torch.Tensor):
    """ subgraph building through masking the unselected nodes with zero features """
    edge_index = data.edge_index
    X = data.x * node_mask.unsqueeze(1)
    if data.pos is None:
        pos = None
    else:
        pos = data.pos * node_mask.unsqueeze(1)
    return X, pos, edge_index


def graph_build_split(data, node_mask: torch.Tensor):
    """ subgraph building through spliting the selected nodes from the original graph """
    x, pos, edge_index = data.x, data.pos, data.edge_index
    row, col = edge_index
    edge_mask = (node_mask[row] == 1) & (node_mask[col] == 1)
    edge_index = edge_index[:, edge_mask]

    sub_nodes = torch.unique(edge_index)
    if sub_nodes.numel() == 0:
        x = torch.zeros_like(data.x)
        if data.pos is not None:
            pos = torch.zeros_like(data.pos)
        return x, pos, edge_index

    x = x[sub_nodes]
    if data.pos is not None:
        pos = data.pos
        pos = pos[sub_nodes]
    else:
        pos = None

    row, col = edge_index
    num_nodes = data.num_nodes
    # remapping the nodes in the explanatory subgraph to new ids.
    node_idx = row.new_full((num_nodes,), -1)
    node_idx[sub_nodes] = torch.arange(sub_nodes.size(0), device=row.device)
    edge_index = node_idx[edge_index]

    return x, pos, edge_index


def l_shapley(coalition: list, data: Data, local_raduis: int,
              value_func: Callable, subgraph_building_method='zero_filling'):
    """ shapley value where players are local neighbor nodes """
    graph = to_networkx(data)
    num_nodes = graph.number_of_nodes()
    subgraph_build_func = get_graph_build_func(subgraph_building_method)

    local_region = copy.copy(coalition)
    for k in range(local_raduis - 1):
        k_neiborhoood = []
        for node in local_region:
            k_neiborhoood += list(graph.neighbors(node))
        local_region += k_neiborhoood
        local_region = list(set(local_region))

    set_exclude_masks = []
    set_include_masks = []
    nodes_around = [node for node in local_region if node not in coalition]
    num_nodes_around = len(nodes_around)

    for subset_len in range(0, num_nodes_around + 1):
        node_exclude_subsets = combinations(nodes_around, subset_len)
        for node_exclude_subset in node_exclude_subsets:
            set_exclude_mask = np.ones(num_nodes)
            set_exclude_mask[local_region] = 0.0
            if node_exclude_subset:
                set_exclude_mask[list(node_exclude_subset)] = 1.0
            set_include_mask = set_exclude_mask.copy()
            set_include_mask[coalition] = 1.0

            set_exclude_masks.append(set_exclude_mask)
            set_include_masks.append(set_include_mask)

    exclude_mask = np.stack(set_exclude_masks, axis=0)
    include_mask = np.stack(set_include_masks, axis=0)
    num_players = len(nodes_around) + 1
    num_player_in_set = num_players - 1 + len(coalition) - (1 - exclude_mask).sum(axis=1)
    p = num_players
    S = num_player_in_set
    coeffs = torch.tensor(1.0 / comb(p, S) / (p - S + 1e-6))

    marginal_contributions = \
        marginal_contribution(data, exclude_mask, include_mask, value_func, subgraph_build_func)

    l_shapley_value = (marginal_contributions.squeeze().cpu() * coeffs).sum().item()
    return l_shapley_value


def mc_shapley(coalition: list, data: Data,
               value_func: Callable, subgraph_building_method='zero_filling',
               sample_num=1000) -> float:
    """ monte carlo sampling approximation of the shapley value """
    subset_build_func = get_graph_build_func(subgraph_building_method)

    num_nodes = data.num_nodes
    node_indices = np.arange(num_nodes)
    coalition_placeholder = num_nodes
    set_exclude_masks = []
    set_include_masks = []

    for example_idx in range(sample_num):
        subset_nodes_from = [node for node in node_indices if node not in coalition]
        random_nodes_permutation = np.array(subset_nodes_from + [coalition_placeholder])
        random_nodes_permutation = np.random.permutation(random_nodes_permutation)
        split_idx = np.where(random_nodes_permutation == coalition_placeholder)[0][0]
        selected_nodes = random_nodes_permutation[:split_idx]
        set_exclude_mask = np.zeros(num_nodes)
        set_exclude_mask[selected_nodes] = 1.0
        set_include_mask = set_exclude_mask.copy()
        set_include_mask[coalition] = 1.0

        set_exclude_masks.append(set_exclude_mask)
        set_include_masks.append(set_include_mask)

    exclude_mask = np.stack(set_exclude_masks, axis=0)
    include_mask = np.stack(set_include_masks, axis=0)
    marginal_contributions = marginal_contribution(data, exclude_mask, include_mask, value_func, subset_build_func)
    mc_shapley_value = marginal_contributions.mean().item()

    return mc_shapley_value


def mc_l_shapley(coalition: list, data: Data, local_raduis: int,
                 value_func: Callable, subgraph_building_method='zero_filling',
                 sample_num=1000) -> float:
    """ monte carlo sampling approximation of the l_shapley value """
    graph = to_networkx(data)
    num_nodes = graph.number_of_nodes()
    subgraph_build_func = get_graph_build_func(subgraph_building_method)

    local_region = copy.copy(coalition)
    for k in range(local_raduis - 1):
        k_neiborhoood = []
        for node in local_region:
            k_neiborhoood += list(graph.neighbors(node))
        local_region += k_neiborhoood
        local_region = list(set(local_region))

    coalition_placeholder = num_nodes
    set_exclude_masks = []
    set_include_masks = []
    for example_idx in range(sample_num):
        subset_nodes_from = [node for node in local_region if node not in coalition]
        random_nodes_permutation = np.array(subset_nodes_from + [coalition_placeholder])
        random_nodes_permutation = np.random.permutation(random_nodes_permutation)
        split_idx = np.where(random_nodes_permutation == coalition_placeholder)[0][0]
        selected_nodes = random_nodes_permutation[:split_idx]

        # if nx.connected_components(graph.subgraph(subgraph_coalition))

        set_exclude_mask = np.ones(num_nodes)
        set_exclude_mask[local_region] = 0.0
        set_exclude_mask[selected_nodes] = 1.0
        set_include_mask = set_exclude_mask.copy()
        set_include_mask[coalition] = 1.0

        set_exclude_masks.append(set_exclude_mask)
        set_include_masks.append(set_include_mask)

    exclude_mask = np.stack(set_exclude_masks, axis=0)
    include_mask = np.stack(set_include_masks, axis=0)
    marginal_contributions = \
        marginal_contribution(data, exclude_mask, include_mask, value_func, subgraph_build_func)

    mc_l_shapley_value = (marginal_contributions).mean().item()
    return mc_l_shapley_value


def gnn_score(coalition: list, data: Data, value_func: Callable,
              subgraph_building_method='zero_filling') -> torch.Tensor:
    """ the value of subgraph with selected nodes """
    num_nodes = data.num_nodes
    subgraph_build_func = get_graph_build_func(subgraph_building_method)
    mask = torch.zeros(num_nodes).type(torch.float32).to(data.x.device)
    mask[coalition] = 1.0
    ret_x, ret_pos, ret_edge_index = subgraph_build_func(data, mask)
    mask_data = Data(x=ret_x, pos=ret_pos, edge_index=ret_edge_index)
    mask_data = Batch.from_data_list([mask_data])
    score = value_func(data) - value_func(mask_data)
    # get the score of predicted class for graph or specific node idx
    return (1-score).item()


def NC_mc_l_shapley(coalition: list, data: Data, local_raduis: int,
                    value_func: Callable, node_idx: int=-1, subgraph_building_method='zero_filling', sample_num=1000) -> float:
    """ monte carlo approximation of l_shapley where the target node is kept in both subgraph """
    graph = to_networkx(data)
    num_nodes = graph.number_of_nodes()
    subgraph_build_func = get_graph_build_func(subgraph_building_method)

    local_region = copy.copy(coalition)
    for k in range(local_raduis - 1):
        k_neiborhoood = []
        for node in local_region:
            k_neiborhoood += list(graph.neighbors(node))
        local_region += k_neiborhoood
        local_region = list(set(local_region))

    coalition_placeholder = num_nodes
    set_exclude_masks = []
    set_include_masks = []
    for example_idx in range(sample_num):
        subset_nodes_from = [node for node in local_region if node not in coalition]
        random_nodes_permutation = np.array(subset_nodes_from + [coalition_placeholder])
        random_nodes_permutation = np.random.permutation(random_nodes_permutation)
        split_idx = np.where(random_nodes_permutation == coalition_placeholder)[0][0]
        selected_nodes = random_nodes_permutation[:split_idx]
        set_exclude_mask = np.ones(num_nodes)
        set_exclude_mask[local_region] = 0.0
        set_exclude_mask[selected_nodes] = 1.0
        if node_idx != -1:
            set_exclude_mask[node_idx] = 1.0
        set_include_mask = set_exclude_mask.copy()
        set_include_mask[coalition] = 1.0  # include the node_idx

        set_exclude_masks.append(set_exclude_mask)
        set_include_masks.append(set_include_mask)

    exclude_mask = np.stack(set_exclude_masks, axis=0)
    include_mask = np.stack(set_include_masks, axis=0)
    marginal_contributions = \
        marginal_contribution(data, exclude_mask, include_mask, value_func, subgraph_build_func)

    mc_l_shapley_value = (marginal_contributions).mean().item()
    return mc_l_shapley_value


import torch
import torch.nn as nn
import numpy as np
import torch.nn.functional as F
from torch_scatter import scatter, scatter_max
from ..base import BaseRandom


class LRIGaussian(BaseRandom):

    def __init__(self, clf, extractor, criterion, config):
        super().__init__()
        self.name = 'lri_gaussian'
        self.clf = clf
        self.extractor = extractor
        self.criterion = criterion
        self.device = next(self.parameters()).device

        self.pred_loss_coef = config['pred_loss_coef']
        self.info_loss_coef = config['info_loss_coef']
        self.dim = config['covar_dim']
        self.attn_constraint = config['attn_constraint']

    @staticmethod
    def kl(pred_sigma, reg_sigma):
        first_term = torch.log(reg_sigma.det() / (abs(pred_sigma.det()) + 1e-6))
        second_term = -reg_sigma.shape[0]
        third_term = batch_trace(torch.inverse(reg_sigma) @ pred_sigma)
        return (first_term + second_term + third_term) / 2

    def __loss__(self, pred_sigma, clf_logits, clf_labels, epoch):
        pred_loss = self.criterion(clf_logits, clf_labels.float())

        reg_sigma = self.get_reg_sigma(epoch)
        info_loss = self.kl(pred_sigma, reg_sigma).mean()

        pred_loss = self.pred_loss_coef * pred_loss
        info_loss = self.info_loss_coef * info_loss

        loss = pred_loss + info_loss
        loss_dict = {'loss': loss.item(), 'pred': pred_loss.item(), 'info': info_loss.item(), 'sig': reg_sigma.det().item()}
        return loss, loss_dict

    def forward(self, data, edge_attn=None):
        original_clf_logits = self.clf(data, edge_attn=edge_attn)
        return original_clf_logits

    def forward_pass(self, data, epoch, do_sampling):
        emb, edge_index = self.clf.get_emb(data)
        U = self.extractor(emb, batch=data.batch, pool_out_lig=None)

        sig1 = F.softplus(U[:, [0]]).clamp(1e-6, 1e6)
        sig2 = F.softplus(U[:, [1]]).clamp(1e-6, 1e6)

        U = U[:, 2:].reshape(emb.shape[0], self.dim, self.dim)
        pred_sigma = sig1.reshape(-1, 1, 1) * U @ U.transpose(1, 2) + sig2.reshape(-1, 1, 1) * torch.eye(self.dim, device=self.device).reshape(-1, self.dim, self.dim)

        if self.attn_constraint == "smooth_min":
           pred_sigma = self.smooth_min(edge_index, pred_sigma, U, sig1, sig2) 
        # pred_sigma = self.smooth_min(edge_index, pred_sigma, U, sig1, sig2) if self.attn_constraint == 'smooth_min' else pred_sigma

        node_noise = self.sampling(U, do_sampling, sig1, sig2)
        masked_clf_logits = self.clf(data, node_noise=node_noise)
        original_clf_logits = self.clf(data)

        loss, loss_dict = self.__loss__(pred_sigma, masked_clf_logits, data.y, epoch)

        return loss, loss_dict, masked_clf_logits, 1-(pred_sigma.det()).reshape(-1)

    def sampling(self, U, do_sampling, sig1, sig2):
        if do_sampling:
            epsilon_1 = torch.randn((U.shape[0], U.shape[2], 1), device=self.device)
            epsilon_2 = torch.randn((U.shape[0], U.shape[2], 1), device=self.device)
            z = sig1.sqrt() * (U @ epsilon_1).squeeze(-1) + sig2.sqrt() * epsilon_2.squeeze(-1)
        else:
            z = None
        return z

    def get_reg_sigma(self, epoch):
        reg_var = torch.tensor([1.0]*self.dim, device=self.device) * 1.0
        reg_sigma = torch.diag_embed(reg_var)
        return reg_sigma

    @staticmethod
    def smooth_min(edge_index, pred_sigma, U, sig1, sig2):
        det = torch.det(pred_sigma)
        _, min_neighbour_idx = scatter_max(det[edge_index[1]], edge_index[0])
        U = U[edge_index[1]][min_neighbour_idx]
        sig1 = sig1[edge_index[1]][min_neighbour_idx]
        sig2 = sig2[edge_index[1]][min_neighbour_idx]
        pred_sigma = pred_sigma[edge_index[1]][min_neighbour_idx]

        det = torch.det(pred_sigma)
        _, min_neighbour_idx = scatter_max(det[edge_index[1]], edge_index[0])
        U = U[edge_index[1]][min_neighbour_idx]
        sig1 = sig1[edge_index[1]][min_neighbour_idx]
        sig2 = sig2[edge_index[1]][min_neighbour_idx]
        pred_sigma = pred_sigma[edge_index[1]][min_neighbour_idx]
        return pred_sigma

def batch_trace(batch_sig):
    return batch_sig.diagonal(offset=0, dim1=-1, dim2=-2).sum(-1)



# https://github.com/vgsatorras/egnn/blob/main/qm9/models.py
# https://github.com/vgsatorras/egnn/blob/main/models/gcl.py

import torch
from torch import nn
import torch.nn.functional as F
from ..utils import FeatEncoder, MLP


class EGNN(nn.Module):
    def __init__(self, x_dim, pos_dim, model_config, feat_info, n_categorical_feat_to_use=-1, n_scalar_feat_to_use=-1, **kwargs):
        super().__init__()
        hidden_size = model_config['hidden_size']
        self.n_layers = model_config['n_layers']

        self.x_dim = x_dim
        self.pos_dim = pos_dim
        self.dropout_p = model_config['dropout_p']
        self.dataset_name = kwargs['aux_info']['dataset_name']
        act_fn = MLP.get_act(model_config['act_type'])()
        norm_type = model_config['norm_type']

        self.node_encoder = FeatEncoder(hidden_size, feat_info['node_categorical_feat'], feat_info['node_scalar_feat'], n_categorical_feat_to_use, n_scalar_feat_to_use)
        self.edge_encoder = FeatEncoder(hidden_size, feat_info['edge_categorical_feat'], feat_info['edge_scalar_feat'])

        edges_in_d = hidden_size if 'acts' in self.dataset_name else 0
        self.convs = nn.ModuleList()
        for _ in range(self.n_layers):
            conv = E_GCL_mask(hidden_size, hidden_size, hidden_size, edges_in_d=edges_in_d, nodes_attr_dim=0, act_fn=act_fn, norm_type=norm_type, recurrent=False, coords_weight=1.0, attention=False)
            self.convs.append(conv)

    def forward(self, x, pos, edge_attr, edge_index, batch, edge_attn=None):
        if self.x_dim == 0 and self.pos_dim != 0:
            feats = pos
        elif self.x_dim != 0 and self.pos_dim == 0:
            feats = x
        elif self.x_dim == 0 and self.pos_dim == 0:
            feats = torch.ones(x.shape[0], 1, device=x.device)
        else:
            feats = torch.cat([x, pos], dim=1)

        x = self.node_encoder(feats)
        edge_attr = self.edge_encoder(edge_attr) if 'acts' in self.dataset_name else None
        for i in range(self.n_layers):
            identity = x
            x, _, _ = self.convs[i](x, edge_index, pos, batch=batch, edge_attr=edge_attr, edge_attn=edge_attn)
            x = x + identity
            x = F.dropout(x, self.dropout_p, training=self.training)
        return x


class E_GCL(nn.Module):
    """Graph Neural Net with global state and fixed number of nodes per graph.
    Args:
          hidden_dim: Number of hidden units.
          num_nodes: Maximum number of nodes (for self-attentive pooling).
          global_agg: Global aggregation function ('attn' or 'sum').
          temp: Softmax temperature.
    """

    def __init__(self, input_nf, output_nf, hidden_nf, edges_in_d=0, nodes_att_dim=0, act_fn=nn.ReLU(), recurrent=True, coords_weight=1.0, attention=False, clamp=False, norm_diff=False, tanh=False):
        super(E_GCL, self).__init__()
        input_edge = input_nf * 2
        self.coords_weight = coords_weight
        self.recurrent = recurrent
        self.attention = attention
        self.norm_diff = norm_diff
        self.tanh = tanh
        edge_coords_nf = 1


        self.edge_mlp = nn.Sequential(
            nn.Linear(input_edge + edge_coords_nf + edges_in_d, hidden_nf),
            act_fn,
            nn.Linear(hidden_nf, hidden_nf),
            act_fn)

        self.node_mlp = nn.Sequential(
            nn.Linear(hidden_nf + input_nf + nodes_att_dim, hidden_nf),
            act_fn,
            nn.Linear(hidden_nf, output_nf))

        layer = nn.Linear(hidden_nf, 1, bias=False)
        torch.nn.init.xavier_uniform_(layer.weight, gain=0.001)

        self.clamp = clamp
        coord_mlp = []
        coord_mlp.append(nn.Linear(hidden_nf, hidden_nf))
        coord_mlp.append(act_fn)
        coord_mlp.append(layer)
        if self.tanh:
            coord_mlp.append(nn.Tanh())
            self.coords_range = nn.Parameter(torch.ones(1))*3
        self.coord_mlp = nn.Sequential(*coord_mlp)


        if self.attention:
            self.att_mlp = nn.Sequential(
                nn.Linear(hidden_nf, 1),
                nn.Sigmoid())

        #if recurrent:
        #    self.gru = nn.GRUCell(hidden_nf, hidden_nf)


    def edge_model(self, source, target, radial, edge_attr):
        if edge_attr is None:  # Unused.
            out = torch.cat([source, target, radial], dim=1)
        else:
            out = torch.cat([source, target, radial, edge_attr], dim=1)
        out = self.edge_mlp(out)
        if self.attention:
            att_val = self.att_mlp(out)
            out = out * att_val
        return out

    def node_model(self, x, edge_index, edge_attr, node_attr):
        row, col = edge_index
        agg = unsorted_segment_sum(edge_attr, row, num_segments=x.size(0))
        if node_attr is not None:
            agg = torch.cat([x, agg, node_attr], dim=1)
        else:
            agg = torch.cat([x, agg], dim=1)
        out = self.node_mlp(agg)
        # if self.recurrent:
        #     out = x + out
        return out, agg

    def coord_model(self, coord, edge_index, coord_diff, edge_feat):
        row, col = edge_index
        trans = coord_diff * self.coord_mlp(edge_feat)
        trans = torch.clamp(trans, min=-100, max=100) #This is never activated but just in case it case it explosed it may save the train
        agg = unsorted_segment_mean(trans, row, num_segments=coord.size(0))
        coord += agg*self.coords_weight
        return coord


    def coord2radial(self, edge_index, coord):
        row, col = edge_index
        coord_diff = coord[row] - coord[col]
        radial = torch.sum((coord_diff)**2, 1).unsqueeze(1)

        if self.norm_diff:
            norm = torch.sqrt(radial) + 1
            coord_diff = coord_diff/(norm)

        return radial, coord_diff

    def forward(self, h, edge_index, coord, edge_attr=None, node_attr=None):
        row, col = edge_index
        radial, coord_diff = self.coord2radial(edge_index, coord)

        edge_feat = self.edge_model(h[row], h[col], radial, edge_attr)
        coord = self.coord_model(coord, edge_index, coord_diff, edge_feat)
        h, agg = self.node_model(h, edge_index, edge_feat, node_attr)
        # coord = self.node_coord_model(h, coord)
        # x = self.node_model(x, edge_index, x[col], u, batch)  # GCN
        return h, coord, edge_attr


class E_GCL_mask(E_GCL):
    """Graph Neural Net with global state and fixed number of nodes per graph.
    Args:
          hidden_dim: Number of hidden units.
          num_nodes: Maximum number of nodes (for self-attentive pooling).
          global_agg: Global aggregation function ('attn' or 'sum').
          temp: Softmax temperature.
    """

    def __init__(self, input_nf, output_nf, hidden_nf, edges_in_d=0, nodes_attr_dim=0, act_fn=nn.ReLU(), norm_type='batch', recurrent=True, coords_weight=1.0, attention=False):
        E_GCL.__init__(self, input_nf, output_nf, hidden_nf, edges_in_d=edges_in_d, nodes_att_dim=nodes_attr_dim, act_fn=act_fn, recurrent=recurrent, coords_weight=coords_weight, attention=attention)

        del self.coord_mlp
        self.act_fn = act_fn
        self.norm = MLP.get_norm(norm_type)(hidden_nf)

    def coord_model(self, coord, edge_index, coord_diff, edge_feat):
        row, col = edge_index
        trans = coord_diff * self.coord_mlp(edge_feat)
        agg = unsorted_segment_sum(trans, row, num_segments=coord.size(0))
        coord += agg*self.coords_weight
        return coord

    def forward(self, h, edge_index, coord, batch, edge_attr=None, node_attr=None, edge_attn=None):
        row, col = edge_index
        radial, coord_diff = self.coord2radial(edge_index, coord)

        edge_feat = self.edge_model(h[row], h[col], radial, edge_attr)

        if edge_attn is not None:
            edge_feat = edge_feat * edge_attn

        h, agg = self.node_model(h, edge_index, edge_feat, node_attr)
        h = self.act_fn(self.norm(h))
        return h, coord, edge_attr


def unsorted_segment_sum(data, segment_ids, num_segments):
    """Custom PyTorch op to replicate TensorFlow's `unsorted_segment_sum`."""
    result_shape = (num_segments, data.size(1))
    result = data.new_full(result_shape, 0)  # Init empty result tensor.
    segment_ids = segment_ids.unsqueeze(-1).expand(-1, data.size(1))
    result.scatter_add_(0, segment_ids, data)
    return result


def unsorted_segment_mean(data, segment_ids, num_segments):
    result_shape = (num_segments, data.size(1))
    segment_ids = segment_ids.unsqueeze(-1).expand(-1, data.size(1))
    result = data.new_full(result_shape, 0)  # Init empty result tensor.
    count = data.new_full(result_shape, 0)
    result.scatter_add_(0, segment_ids, data)
    count.scatter_add_(0, segment_ids, torch.ones_like(data))
    return result / count.clamp(min=1)



import yaml
import json
import argparse
import uuid
from tqdm import tqdm
from pathlib import Path
import torch
import networkx as nx
import torch.nn as nn
from torch.nn import functional as F
from eval import FidelEvaluation, AUCEvaluation
from get_model import Model
from baselines import LabelPerturb, VGIB, LRIBern, LRIGaussian, CIGA
from utils import to_cpu, log_epoch, get_data_loaders, set_seed, load_checkpoint, ExtractorMLP, get_optimizer
from utils import inherent_models, post_hoc_explainers, post_hoc_attribution, name_mapping
import torchmetrics
from torch_geometric.data import Data
from torch_geometric.loader import DataLoader
from torch_geometric.utils import subgraph, to_networkx
from rdkit import Chem
from torch_geometric.nn import knn_graph
import numpy as np
import pandas as pd
import warnings
from eval import control_sparsity
from datetime import datetime
import matplotlib.pyplot as plt
warnings.filterwarnings("ignore")


def node_attn_to_edge_attn(node_attn, edge_index):
    src_attn = node_attn[edge_index[0]]
    dst_attn = node_attn[edge_index[1]]
    edge_attn = src_attn * dst_attn
    return edge_attn


def eval_one_sample(baseline, data, epoch):
    with torch.set_grad_enabled(baseline.name in ['gradcam', 'gradgeo', 'gnnexplainer']):
        baseline.extractor.eval() if hasattr(baseline, 'extractor') else None
        baseline.clf.eval()

        # do_sampling = True if phase == 'valid' and baseline.name == 'pgexplainer' else False # we find this is better for BernMaskP
        loss, loss_dict, infer_clf_logits, node_attn = baseline.forward_pass(data, epoch=epoch, do_sampling=False)
        # edge_attn = node_attn_to_edge_attn(node_attn, data.edge_index) if baseline.name != 'lri_gaussian' else None

        return loss_dict, to_cpu(infer_clf_logits), to_cpu(node_attn)  #, to_cpu(edge_attn)

def get_viz_idx(test_set, dataset_name, num_viz_samples):
    y_dist = test_set.data.y.numpy().reshape(-1)
    num_nodes = np.array([each.x.shape[0] for each in test_set])
    classes = np.unique(y_dist)
    res = []
    for each_class in classes:
        tag = 'class_' + str(each_class)
        if dataset_name == 'Graph-SST2':
            condi = (y_dist == each_class) * (num_nodes > 5) * (num_nodes < 10)  # in case too short or too long
            candidate_set = np.nonzero(condi)[0]
        else:
            candidate_set = np.nonzero(y_dist == each_class)[0]
        idx = np.random.choice(candidate_set, num_viz_samples, replace=False)
        res.append((idx, tag))

    if dataset_name == 'synmol':
        for each_class in classes:
            tag = 'class_' + str(each_class)
            candidate_set = np.nonzero(y_dist == each_class)[0]
            idx = np.random.choice(candidate_set, num_viz_samples, replace=False)
            res.append((idx, tag))
    return res

def visualize_results(gsat, test_set, num_viz_samples, dataset_name):
    all_viz_set = get_viz_idx(test_set, num_viz_samples, dataset_name)
    figsize = 10
    fig, axes = plt.subplots(len(all_viz_set), num_viz_samples, figsize=(figsize*num_viz_samples, figsize*len(all_viz_set)*0.8))

    for class_idx, (idx, tag) in enumerate(all_viz_set):
        viz_set = test_set[idx]
        data = next(iter(DataLoader(viz_set, batch_size=len(idx), shuffle=False)))
        batch_att, _, _ = eval_one_batch(gsat, data.to(gsat.device), epoch=500)

        # iterate each graph
        for i in tqdm(range(len(viz_set))):
            mol_type, coor = None, None
            if dataset_name == 'synmol':
                ATOM_TYPES = ['C', 'N', 'O', 'S', 'F', 'P', 'Cl', 'Br', 'Na', 'Ca', 'I', 'B', 'H', '*']
                node_dict = dict(zip(list(range(len(ATOM_TYPES))), ATOM_TYPES))
                mol_type = {k: node_dict[v.item()] for k, v in enumerate(viz_set[i].node_type)}
            else:
                raise NotImplementedError

            # select one graph
            node_subset = data.batch == i
            _, edge_mask = subgraph(node_subset, data.edge_index, edge_attr=batch_att)

            node_label = viz_set[i].node_label.reshape(-1) if viz_set[i].get('node_label', None) is not None else torch.zeros(viz_set[i].x.shape[0])
            visualize_a_graph(viz_set[i].edge_index, edge_mask, node_label, node_type, axes[class_idx, i], norm=True, mol_type=mol_type, coor=coor)
            # axes[class_idx, i].axis('off')
        fig.tight_layout()

    each_plot_len = 1/len(viz_set)
    for num in range(1, len(viz_set)):
        line = plt.Line2D((each_plot_len*num, each_plot_len*num), (0, 1), color="gray", linewidth=1, linestyle='dashed', dashes=(5, 10))
        fig.add_artist(line)

    each_plot_width = 1/len(all_viz_set)
    for num in range(1, len(all_viz_set)):
        line = plt.Line2D((0, 1), (each_plot_width*num, each_plot_width*num), color="gray", linestyle='dashed', dashes=(5, 10))
        fig.add_artist(line)


def visualize_a_graph(edge_index, edge_att, node_label, node_type, ax, coor=None, norm=False, mol_type=None, nodesize=300):
    # if dataset_name == 'synmol':
    ATOM_TYPES = ['C', 'N', 'O', 'S', 'F', 'P', 'Cl', 'Br', 'Na', 'Ca', 'I', 'B', 'H', '*']
    node_dict = dict(zip(list(range(len(ATOM_TYPES))), ATOM_TYPES))
    mol_type = {k: node_dict[v.item()] for k, v in enumerate(node_type)}

    if norm:  # for better visualization
        edge_att = edge_att**10
        edge_att = (edge_att - edge_att.min()) / (edge_att.max() - edge_att.min() + 1e-6)

    if mol_type is None or dataset_name == 'Graph-SST2':
        atom_colors = {0: '#E49D1C', 1: '#FF5357', 2: '#a1c569', 3: '#69c5ba'}
        node_colors = [None for _ in range(node_label.shape[0])]
        for y_idx in range(node_label.shape[0]):
            node_colors[y_idx] = atom_colors[node_label[y_idx].int().tolist()]
    else:
        node_color = ['#29A329', 'lime', '#F0EA00',  'maroon', 'brown', '#E49D1C', '#4970C6', '#FF5357']
        element_idxs = {k: Chem.PeriodicTable.GetAtomicNumber(Chem.GetPeriodicTable(), v) for k, v in mol_type.items()}
        node_colors = [node_color[(v - 1) % len(node_color)] for k, v in element_idxs.items()]

    data = Data(edge_index=edge_index, att=edge_att, y=node_label, num_nodes=node_label.size(0)).to('cpu')
    G = to_networkx(data, node_attrs=['y'], edge_attrs=['att'])

    # calculate Graph positions
    if coor is None:
        pos = nx.kamada_kawai_layout(G)
    else:
        pos = {idx: each.tolist() for idx, each in enumerate(coor)}

    for source, target, data in G.edges(data=True):
        ax.annotate(
            '', xy=pos[target], xycoords='data', xytext=pos[source],
            textcoords='data', arrowprops=dict(
                arrowstyle="->" if dataset_name == 'Graph-SST2' else '-',
                lw=max(data['att'], 0) * 3,
                alpha=max(data['att'], 0),  # alpha control transparency
                color='black',  # color control color
                shrinkA=np.sqrt(nodesize) / 2.0 + 1,
                shrinkB=np.sqrt(nodesize) / 2.0 + 1,
                connectionstyle='arc3,rad=0.4' if dataset_name == 'Graph-SST2' else 'arc3'
            ))

    if mol_type is not None:
        nx.draw_networkx_labels(G, pos, mol_type, ax=ax)

    if dataset_name != 'Graph-SST2':
        nx.draw_networkx_nodes(G, pos, node_color=node_colors, node_size=nodesize, ax=ax)
        nx.draw_networkx_edges(G, pos, width=1, edge_color='gray', arrows=False, alpha=0.1, ax=ax)
    else:
        nx.draw_networkx_edges(G, pos, width=1, edge_color='gray', arrows=False, alpha=0.1, ax=ax, connectionstyle='arc3,rad=0.4')

def get_edges_from_smiles(smiles):
    mol = Chem.MolFromSmiles(smiles.values[0])

    if len(mol.GetBonds()) > 0:  # mol has bonds
        edges_list = []
        for bond in mol.GetBonds():
            i = bond.GetBeginAtomIdx()
            j = bond.GetEndAtomIdx()

            # add edges in both directions
            edges_list.append((i, j))
            edges_list.append((j, i))
        # data.edge_index: Graph connectivity in COO format with shape [2, num_edges]
        edge_index = np.array(edges_list, dtype=np.int64).T
    else:  # mol has no bonds
        edge_index = np.empty((2, 0), dtype=np.int64)
    return torch.tensor(edge_index)

def eval_one_epoch(baseline, data_loader, epoch, phase, seed, signal_class, writer=None, eval_metric=None, metric_list=None):
    view_mode = 0

    pbar, avg_loss_dict = tqdm(data_loader), dict()
    clf_ACC, clf_AUC = torchmetrics.Accuracy(task='binary'), torchmetrics.AUROC(task='binary')
    mis, neg, pos, masked_perf_drop = 0, 0, 0, 0
    mol_df = pd.read_csv('../data/synmol/raw' + '/logic8_smiles.csv')
    for idx, data in enumerate(pbar):
        # data = negative_augmentation(data, data_config, phase, data_loader, idx, loader_len)
        loss_dict, clf_logits, attn = eval_one_sample(baseline, data.to(baseline.device), epoch)
        ex_labels, clf_labels, data = to_cpu(data.node_label), to_cpu(data.y), data.cpu()
        label, pred = int(data.y.item()), int(clf_logits.sigmoid() > 0.5)
        neg += 1 if label == 0 else 0
        pos += 1 if label == 1 else 0
        if label == 1:
            pass
        #     fig, ax = plt.subplots(figsize=(10, 8))
        #     edge_index = get_edges_from_smiles(mol_df.iloc[data.mol_df_idx]['smiles'])
        #     edge_attn = node_attn_to_edge_attn(data.node_label, edge_index)
        #     visualize_a_graph(edge_index, edge_attn, data.node_label, data.x, ax, coor=None, norm=False,
        #                       mol_type=None, nodesize=300)
        #     plt.show()
        #     place = 0
        if label == view_mode:
            [metric.collect_batch(ex_labels, attn, data, signal_class, 'geometric') for metric in metric_list]
            masked_perf_drop = eval_metric.collect_batch(ex_labels, attn, data, signal_class, 'geometric')
            if masked_perf_drop == 1 - view_mode:  # still_pos (view_mode=1 and masked_perf_drop=0) or mis_neg
                mis += 1
                fig, ax = plt.subplots(figsize=(10, 8))
                sparse_attn = control_sparsity(attn, sparsity=eval_metric.sparsity)
                edge_index = get_edges_from_smiles(mol_df.iloc[data.mol_df_idx]['smiles'])
                # edge_index = knn_graph(data.pos, k=2, batch=data.batch, loop=False)
                edge_attn = node_attn_to_edge_attn(sparse_attn, edge_index) if baseline.name != 'lri_gaussian' else None
                visualize_a_graph(edge_index, 1 - edge_attn, data.node_label, data.x, ax, coor=None, norm=False, mol_type=None, nodesize=300)
                # fig.tight_layout()
                masked_pred = 1 - pred if masked_perf_drop == 1 else pred
                plt.title(f'label: {label}, pred: {pred}, masked_pred: {masked_pred}, sparsity: {eval_metric.sparsity}')
                # uid = uuid.uuid1()
                plt.savefig(f'../../img/erm/' + str(data.mol_df_idx.item()) + '.png')
                # plt.show()
        # print()
        eval_dict = {}
        eval_dict.update({'clf_acc': clf_ACC(clf_logits, clf_labels), 'clf_auc': clf_AUC(clf_logits, clf_labels)})
        # eval_dict.update({'mean_fid': mean([eval_dict[k] for k in eval_dict if 'fid' in k])})

        desc = log_epoch(seed, epoch, phase, loss_dict, eval_dict, phase_info='dataset')
        pbar.set_description(desc)
        # compute the avg_loss for the epoch desc
        exec('for k, v in loss_dict.items():\n\tavg_loss_dict[k]=(avg_loss_dict.get(k, 0) * idx + v) / (idx + 1)')
    # print(f"mis_neg/neg/all: {mis}/{neg}/{len(pbar)}")
    epoch_dict = {eval_metric.name: eval_metric.eval_epoch()}
    epoch_dict.update({metric.name: metric.eval_epoch() for metric in metric_list})
    epoch_dict.update({'clf_acc': clf_ACC.compute().item(), 'clf_auc': clf_AUC.compute().item()})
    # epoch_dict.update({'mean_fid': mean([epoch_dict[k] for k in epoch_dict if 'fid' in k])})
    # log_epoch(seed, epoch, phase, avg_loss_dict, epoch_dict, writer)

    return epoch_dict

def test(config, method_name, exp_method, model_name, backbone_seed, dataset_name, log_dir, device):
    set_seed(backbone_seed, method_name=method_name, backbone_name=model_name)
    writer = None
    print('The logging directory is', log_dir), print('=' * 80)

    data_config = config['data']
    loaders, test_set, dataset = get_data_loaders(dataset_name, 1, data_config, dataset_seed=0)
    signal_class = dataset.signal_class

    clf = Model(model_name, config[model_name],  # backbone_config
                method_name, config[method_name],  # method_config
                dataset).to(device)

    extractor = ExtractorMLP(config[model_name]['hidden_size'], config[method_name], config['data'].get('use_lig_info', False)) \
        if method_name in inherent_models + ['pgexplainer'] else nn.Identity()
    extractor = extractor.to(device)
    criterion = F.binary_cross_entropy_with_logits

    # establish the model and the metrics

    backbone = eval(name_mapping[method_name])(clf, extractor, criterion, config[method_name]) \
        if method_name in inherent_models else clf
    assert load_checkpoint(backbone, log_dir, model_name=method_name, seed=backbone_seed, map_location=torch.device('cpu') if not torch.cuda.is_available() else None)

    if 'label_perturb' in exp_method:
        model = LabelPerturb(backbone, mode=int(exp_method[-1]))
    else:
        model = eval(name_mapping[exp_method])(clf, criterion, config[exp_method])

    # metric_list = [FidelEvaluation(backbone, i/10) for i in range(2, 9)]
    main_metric = FidelEvaluation(backbone, 0.8)
    metric_list = [FidelEvaluation(backbone, i/10) for i in range(2, 9)]
    print('Use random explanation and fidelity w/ signal nodes to test the Model Sensitivity.')
    # test_set =
    # DataLoader(test_set, batch_size=1, shuffle=False)
    train_dataset = loaders['train']  # [:100]
    valid_dataset = loaders['valid']  # [:100]
    test_dataset = loaders['test']
    # metric_names = [i.name for i in metric_list] + ['clf_acc', 'clf_auc']
    # train_dict = eval_one_epoch(model, train_dataset, 1, 'train', backbone_seed, signal_class, writer, metric_list)
    # valid_dict = eval_one_epoch(model, valid_dataset, 1, 'valid', backbone_seed, signal_class, writer, metric_list)
    test_dict = eval_one_epoch(model, test_dataset, 1, 'test', backbone_seed, signal_class,  writer, main_metric, metric_list)

    return test_dict

def save_multi_bseeds(multi_methods_res, method_name, exp_method, seeds):
    seeds += ['avg', 'std']
    indexes = [method_name+'_'+str(seed) for seed in seeds]
    # from itertools import product
    # indexes = ['_'.join(item) for item in product(methods, seeds)]
    df = pd.DataFrame(multi_methods_res, index=indexes)

    day_dir = Path('result') / config_name / datetime.now().strftime("%m_%d")
    day_dir.mkdir(parents=True, exist_ok=True)
    csv_dir = day_dir / ('_'.join([exp_method, method_name, 'sensitivity.csv']))
    with open(csv_dir, mode='w') as f:
        df.to_csv(f, lineterminator="\n", header=f.tell() == 0)
    return df

def get_avg_std_report(reports):
    # print(reports)
    all_keys = {k: [] for k in reports[0]}
    for report in reports:
        for k in report:
            all_keys[k].append(report[k])
    avg_report = {k: np.mean(v) for k, v in all_keys.items()}
    std_report = {k: np.std(v) for k, v in all_keys.items()}
    avg_std_report = {k: f'{np.mean(v):.3f} +/- {np.std(v):.3f}' for k, v in all_keys.items()}
    return avg_report, std_report, avg_std_report

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='EXP')
    parser.add_argument('-d', '--dataset', type=str, help='dataset used', default='synmol')
    parser.add_argument('--clf_method', type=str, help='method used', default='erm', choices=inherent_models+['erm'])
    parser.add_argument('--exp_method', type=str, help='method used', default='label_perturb1', choices=['label_perturb1', 'label_perturb0'])
    parser.add_argument('-b', '--backbone', type=str, help='backbone used', default='egnn')
    parser.add_argument('--cuda', type=int, help='cuda device id, -1 for cpu', default=-1)
    parser.add_argument('--note', type=str, help='note in log name', default='')
    parser.add_argument('--bseeds', type=int, nargs="+", help='random seed for training backbone', default=[0, 1, 2, 3, 4, 5, 6, 7, 8, 9])
    # parser.add_argument('--mseed', type=int, help='random seed for explainer', default=0)

    args = parser.parse_args()
    # sub_metric = 'avg_loss'
    print(args)

    dataset_name, method_name, model_name, cuda_id, note = args.dataset, args.clf_method, args.backbone, args.cuda, args.note
    exp_method = args.exp_method
    config_name = '_'.join([model_name, dataset_name])
    # sub_dataset_name = '_' + dataset_name.split('_')[1] if len(dataset_name.split('_')) > 1 else ''
    config_path = Path('./configs') / f'{config_name}.yml'
    config = yaml.safe_load((config_path).open('r'))
    if config[method_name].get(model_name, False):
        config[method_name].update(config[method_name][model_name])

    print('=' * 80)
    print(f'Config for {method_name}: ', json.dumps(config[method_name], indent=4))



    # the directory is like egnn_actstrack / bseed0 / lri_bern
    multi_seeds_res = []
    for backbone_seed in args.bseeds:
        device = torch.device(f'cuda:{cuda_id}' if cuda_id >= 0 and torch.cuda.is_available() else 'cpu')
        model_dir = Path('log') / config_name / method_name
        test_report = test(config, method_name, exp_method, model_name, backbone_seed, dataset_name, model_dir, device)
        multi_seeds_res += [test_report]
        # print('Train Dataset Result: ', json.dumps(train_report, indent=4))
        # print('Valid Dataset Result: ', json.dumps(valid_report, indent=4))
        print('Test Dataset Result: ', json.dumps(test_report, indent=4))
    avg_report, std_report, avg_std_report = get_avg_std_report(multi_seeds_res)
    multi_seeds_res += [avg_report, std_report]
    print(json.dumps(avg_std_report, indent=4))
    save_multi_bseeds(multi_seeds_res, method_name, exp_method, args.bseeds)




# https://github.com/jacobgil/pytorch-grad-cam/blob/master/pytorch_grad_cam/base_cam.py

import numpy as np
import torch
import torch.nn as nn
from ..base import BaseRandom


class Grad(BaseRandom):
    def __init__(self, clf, signal_class, criterion, config) -> None:
        super().__init__()
        self.clf = clf
        self.target_layers = [clf.model.convs[-1]]
        self.criterion = criterion
        self.gradgeo = config.get('gradgeo', False)
        self.device = next(self.parameters()).device

        assert signal_class is not None
        self.signal_class = signal_class

    def start_tracking(self):
        self.activations_and_grads = ActivationsAndGradients(self.clf, self.target_layers)

    def get_cam(self, activations: torch.Tensor, grads: torch.Tensor) -> np.ndarray:
        assert activations.min() >= 0.0  # rectified
        cam = (grads * activations).sum(1)
        return cam

    def aggregate_multi_layers(self, cam_per_target_layer: np.ndarray) -> np.ndarray:
        cam_per_target_layer = np.stack(cam_per_target_layer, axis=1)
        result = np.mean(cam_per_target_layer, axis=1)
        return result

    def __loss__(self, attn, clf_logits, clf_labels, epoch, warmup):
        assert warmup
        pred_loss = self.criterion(clf_logits, clf_labels.float())
        if warmup:
            return pred_loss, {'loss': pred_loss.item(), 'pred': pred_loss.item()}

    def forward_pass(self, data, epoch, warmup, **kwargs):
        if warmup:
            clf_logits = self.clf(data)
            loss, loss_dict = self.__loss__(None, clf_logits, data.y, epoch, warmup)
            return loss, loss_dict, clf_logits, None, None, None, None

        self.clf.eval()
        if self.gradgeo:
            data.pos.requires_grad = True

        original_clf_logits = self.activations_and_grads(data)
        masked_clf_logits = original_clf_logits
        targets = [BinaryClassifierOutputTarget(self.signal_class)] * original_clf_logits.shape[0]
        for k in self.clf.model.state_dict():
            print(k)

        self.clf.zero_grad()
        loss = sum([target(output) for target, output in zip(targets, original_clf_logits)])
        loss.backward(retain_graph=True)
        loss_dict = {'loss': loss.item(), 'pred': loss.item()}

        pred_dir = None
        if self.gradgeo:
            node_weights = data.pos.grad.norm(dim=1, p=2)
            pred_dir = data.pos.grad[:, :2]
        else:
            cam_per_layer = self.compute_cam_per_layer()
            node_weights = torch.tensor(self.aggregate_multi_layers(cam_per_layer))
        return loss, loss_dict, original_clf_logits, masked_clf_logits, node_weights.reshape(-1), pred_dir, None

    def compute_cam_per_layer(self) -> np.ndarray:
        activations_list = [a.cpu().data.numpy() for a in self.activations_and_grads.activations]
        grads_list = [g.cpu().data.numpy() for g in self.activations_and_grads.gradients]

        cam_per_target_layer = []
        # Loop over the saliency image from every layer
        for i in range(len(self.target_layers)):
            layer_activations = None
            layer_grads = None
            if i < len(activations_list):
                layer_activations = activations_list[i]
            if i < len(grads_list):
                layer_grads = grads_list[i]

            cam = self.get_cam(layer_activations, layer_grads)
            cam_per_target_layer.append(cam)
        return cam_per_target_layer

    def __del__(self):
        self.activations_and_grads.release()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        self.activations_and_grads.release()
        if isinstance(exc_value, IndexError):
            # Handle IndexError here...
            print(
                f"An exception occurred in CAM with block: {exc_type}. Message: {exc_value}")
            return True


class BinaryClassifierOutputTarget:
    def __init__(self, category):
        self.category = category

    def __call__(self, model_output):
        if self.category == 1:
            sign = 1
        else:
            sign = -1
        return model_output * sign


class ActivationsAndGradients:
    """ Class for extracting activations and
    registering gradients from targetted intermediate layers """

    def __init__(self, model, target_layers, reshape_transform=None):
        self.model = model
        self.gradients = []
        self.activations = []
        self.reshape_transform = reshape_transform
        self.handles = []
        for target_layer in target_layers:
            self.handles.append(
                target_layer.register_forward_hook(self.save_activation))
            # Because of https://github.com/pytorch/pytorch/issues/61519,
            # we don't use backward hook to record gradients.
            self.handles.append(
                target_layer.register_forward_hook(self.save_gradient))

    def save_activation(self, module, input, output):
        if isinstance(output, tuple):  # egnn outputs a tuple
            output = output[0]

        activation = output

        if self.reshape_transform is not None:
            activation = self.reshape_transform(activation)
        self.activations.append(activation.cpu().detach())

    def save_gradient(self, module, input, output):
        if isinstance(output, tuple):  # egnn outputs a tuple
            output = output[0]

        if not hasattr(output, "requires_grad") or not output.requires_grad:
            # You can only register hooks on tensor requires grad.
            return

        # Gradients are computed in reverse order
        def _store_grad(grad):
            if self.reshape_transform is not None:
                grad = self.reshape_transform(grad)
            self.gradients = [grad.cpu().detach()] + self.gradients

        output.register_hook(_store_grad)

    def __call__(self, x):
        self.gradients = []
        self.activations = []
        return self.model(x)

    def release(self):
        for handle in self.handles:
            handle.remove()



import os
pre_cuda_id = int(os.environ.get('CUDA_VISIBLE_DEVICES')) if os.environ.get('CUDA_VISIBLE_DEVICES') else None
# os.environ['CUDA_VISIBLE_DEVICES'] = '0,1,2,3,4,5'
import yaml
import json
import pandas as pd
import argparse
from tqdm import tqdm
from pathlib import Path
from copy import deepcopy
from itertools import product
import nni
import re
import torch
import torch.nn as nn
from torch.nn import functional as F
from torch.utils.tensorboard import SummaryWriter
from eval import FidelEvaluation, LabelFidelity, AUCEvaluation
from get_model import Model
from baselines import *
from utils import to_cpu, log_epoch, get_data_loaders, set_seed, update_and_save_best_epoch_res, load_checkpoint, save_checkpoint, ExtractorMLP, get_optimizer, sum_fid_score
from utils import inherent_models, post_hoc_explainers, post_hoc_attribution, name_mapping
import torchmetrics
from statistics import mean
import warnings
warnings.filterwarnings("ignore")


def eval_one_batch(baseline, data, epoch, phase):
    with torch.set_grad_enabled(baseline.name in ['gradcam', 'gradx', 'inter_grad', 'gnnexplainer']):
        baseline.extractor.eval() if hasattr(baseline, 'extractor') else None
        baseline.clf.eval()

        # BernMaskP
        do_sampling = True if phase == 'valid' and baseline.name == 'pgexplainer' else False # we find this is better for BernMaskP
        loss, loss_dict, infer_clf_logits, node_attn = baseline.forward_pass(data, epoch=epoch, do_sampling=do_sampling)

        return loss_dict, to_cpu(infer_clf_logits), to_cpu(node_attn)


def test_one_epoch(baseline, data_loader, epoch, phase, seed, signal_class, writer=None, metric_list=None):
    use_tqdm = True
    run_one_batch = eval_one_batch
    pbar, avg_loss_dict = tqdm(data_loader) if use_tqdm else data_loader, dict()
    [eval_metric.reset() for eval_metric in metric_list] if metric_list else None

    clf_ACC, clf_AUC = torchmetrics.Accuracy(task='binary'), torchmetrics.AUROC(task='binary')
    save_epoch_attn = []
    for idx, data in enumerate(pbar):
        # data = negative_augmentation(data, data_config, phase, data_loader, idx, loader_len)
        loss_dict, clf_logits, attn = run_one_batch(baseline, data.to(baseline.device), epoch, phase)
        ex_labels, clf_labels, data = to_cpu(data.node_label), to_cpu(data.y), data.cpu()


        eval_dict = {metric.name: metric.collect_batch(ex_labels, attn, data, signal_class, 'geometric')
                     for metric in metric_list} if metric_list else {}
        eval_dict.update({'clf_acc': clf_ACC(clf_logits, clf_labels), 'clf_auc': clf_AUC(clf_logits, clf_labels)})
        batch_fid = [eval_dict[k] for k in eval_dict if 'fid' in k and 'all' in k]
        eval_dict.update({'fid_all': mean(batch_fid)}) if phase in ['valid', 'test'] and batch_fid else {}


        desc = log_epoch(seed, epoch, phase, loss_dict, eval_dict)
        pbar.set_description(desc) if use_tqdm else None
        # compute the avg_loss for the epoch desc
        exec('for k, v in loss_dict.items():\n\tavg_loss_dict[k]=(avg_loss_dict.get(k, 0) * idx + v) / (idx + 1)')

    epoch_dict = {eval_metric.name: eval_metric.eval_epoch() for eval_metric in metric_list} if metric_list else {}
    epoch_dict.update({'clf_acc': clf_ACC.compute().item(), 'clf_auc': clf_AUC.compute().item()})
    sum_fid_score(epoch_dict)

    log_epoch(seed, epoch, phase, avg_loss_dict, epoch_dict, writer)

    return epoch_dict


def test(config, method_name, model_name, backbone_seed, seed, dataset_name, parent_dir, device, quick=False, load_dir=None):
    # writer = SummaryWriter(log_dir) if log_dir is not None else None
    writer = None
    model_dir, log_dir = (parent_dir / 'erm', None) if model_name in post_hoc_attribution \
        else (parent_dir / 'erm', parent_dir / method_name)
    # log_dir = parent_dir / method_name if method_name in inherent_models else None
    # log_dir.mkdir(parents=True, exist_ok=True) if log_dir is not None else None
    # model_dir.mkdir(parents=True, exist_ok=True)

    batch_size = config['optimizer']['batch_size']
    model_cofig = config[method_name] if method_name in inherent_models else config['erm']
    warmup = model_cofig['warmup']
    epochs = config[method_name]['epochs']
    data_config = config['data']
    loaders, test_set, dataset = get_data_loaders(dataset_name, batch_size, data_config, dataset_seed=0, num_workers=0, device=device)
    signal_class = dataset.signal_class

    clf = Model(model_name, config[model_name],  # backbone_config
                method_name, config[method_name],  # method_config
                dataset).to(device)
    extractor = ExtractorMLP(config[model_name]['hidden_size'], config[method_name], config['data'].get('use_lig_info', False)) \
        if method_name in inherent_models + ['pgexplainer'] else nn.Identity()
    extractor = extractor.to(device)
    criterion = F.binary_cross_entropy_with_logits
    constructor = eval(name_mapping[method_name])

    # establish the model and the metrics
    assert method_name in post_hoc_attribution + post_hoc_explainers
    baseline = constructor(clf, criterion, config[method_name]) if method_name != 'pgexplainer' else PGExplainer(clf, extractor, criterion, config['pgexplainer'])
    # assert load_checkpoint(baseline.clf, model_dir, model_name='erm', seed=backbone_seed, map_location=)
    map_location = torch.device('cpu') if not torch.cuda.is_available() else device
    checkpoint = torch.load(load_dir, map_location=map_location)
    baseline.clf.load_state_dict(checkpoint['model_state_dict'])

        # save_checkpoint(baseline.clf, model_dir, model_name='erm', backbone_seed=backbone_seed, seed=backbone_seed)
    metric_list = [AUCEvaluation()] #+ [FidelEvaluation(baseline.clf, i/10) for i in range(2, 9)] + \
              # [FidelEvaluation(baseline.clf, i/10, instance='pos') for i in range(2, 9)] + \
              # [FidelEvaluation(baseline.clf, i/10, instance='neg') for i in range(2, 9)] if quick==False else \
              # [AUCEvaluation()] + [FidelEvaluation(baseline.clf, i/10) for i in range(2, 9)]
    baseline.start_tracking() if 'grad' in method_name or method_name == 'gnnlrp' else None

    set_seed(seed, method_name=method_name, backbone_name=model_name)
    metric_names = [a + b for a, b in product(['valid_', 'test_'], [i.name for i in metric_list]+['clf_acc', 'clf_auc', 'fid_all', 'fid_pos', 'fid_neg'])]
    # metric_names = [j+i.name for i in metric_list for j in ['valid_', 'test_']]
    metric_dict = {}.fromkeys(metric_names, -1)
    assert epochs == 1 # methods only one epoch
    test_dict = test_one_epoch(baseline, loaders['test'], 1, 'test', seed,
                                              signal_class,  writer, metric_list)
        # metric_dict, new_best = update_and_save_best_epoch_res(baseline, metric_dict, valid_dict, test_dict, epoch, log_dir, backbone_seed, seed, writer, method_name, main_metric)

    return test_dict


def test_one_seed(args, load_dir):
    print(args)
    dataset_name, method_name, model_name, cuda_id, note = args.dataset, args.method, args.backbone, args.cuda, args.note
    method_seed, backbone_seed = args.seed, args.bseed
    set_seed(backbone_seed, method_name=method_name, backbone_name=model_name)
    config_name = '_'.join([model_name, dataset_name])
    # sub_dataset_name = '_' + dataset_name.split('_')[1] if len(dataset_name.split('_')) > 1 else ''
    config_path = Path('./configs') / f'{config_name}.yml'
    config = yaml.safe_load((config_path).open('r'))
    if config[method_name].get(model_name, False):
        config[method_name].update(config[method_name][model_name])
    print('=' * 80)
    print(f'Config for {method_name}: ', json.dumps(config[method_name], indent=4))
    # :{pre_cuda_id}
    cuda_ = f'cuda' if cuda_id == 99 else f'cuda:{cuda_id}' if cuda_id >= 0 else 'cpu'
    device = torch.device(cuda_)
    # log_dir = None
    # if config['logging']['tensorboard'] or method_name in ['gradcam', 'gradgeo', 'bernmask_p', 'bernmask']:
    main_dir = Path('log') / config_name
    main_dir.mkdir(parents=True, exist_ok=True)
        # shutil.copy(config_path, log_dir / config_path.name)
    report_dict = test(config, method_name, model_name, backbone_seed, method_seed, dataset_name, main_dir, device, quick=args.quick, load_dir=load_dir)
    # want_keys = [a + b for a in ['valid_', 'test_'] for b in ['fid_all', 'fid_pos', 'fid_neg', 'exp_auc', 'clf_auc', 'clf_acc']]
    # sub_report = {k: v for k, v in report_dict.items() if k in want_keys}
    # print(sub_report)
    return report_dict


def main(args):
    # if args.gpu_ratio is not None:
    #     torch.cuda.set_per_process_memory_fraction(args.gpu_ratio)
    report_dict, indexes = [], []

    def findAllFile(base, restrict=None):
        res = []
        for root, ds, fs in os.walk(base):
            for f in fs:
                if not restrict or restrict in f:
                    fullname = os.path.join(root, f)
                    res += [fullname]
        return res

    def findSubFile(base, restrict=None):
        res = []
        for f in os.listdir(base):
            if not restrict or restrict in f:
                fullname = os.path.join(base, f)
                res += [fullname]
        return res

    config_name = '_'.join([args.backbone, args.dataset])
    for model_path in findSubFile(Path('log') / config_name / 'erm' / 'models', restrict='pt'):
        # args.method = method
        path_list = re.split(r'[\./-]', model_path) # /seed0-9epo-999auc.pt
        auc_name, seed_name = path_list[-2], path_list[-4] # the former five letters are seed0_
        name = auc_name + f'_{seed_name[-1]}'
        print(f'The model to load is {name}. ({path_list[-3]})')
        report_dict += [test_one_seed(args, load_dir=model_path)]
        indexes += [name]
    # print(json.dumps(report_dict, indent=4))

    df_path = Path('result') / config_name / 'model_variation'
    df_path.mkdir(parents=True, exist_ok=True)
    df = pd.DataFrame(report_dict, index=indexes)
    df.to_csv(df_path / f'{args.method}.csv')
    # nni_report = dict(report_dict, **{'default': report_dict[f'valid_{main_metric}']})

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='EXP')
    parser.add_argument('-d', '--dataset', type=str, help='dataset used', default='actstrack_2T')
    parser.add_argument('-m', '--method', type=str, help='method used', default='')
    # parser.add_argument('--methods', type=str, nargs='+', help='methods used', default='')
    parser.add_argument('-b', '--backbone', type=str, help='backbone used', default='egnn')
    parser.add_argument('--cuda', type=int, help='cuda device id, -1 for cpu', default=-1)
    parser.add_argument('--seed', type=int, help='random seed', default=0)
    parser.add_argument('--note', type=str, help='note in log name', default='')
    parser.add_argument('--gpu_ratio', type=float, help='gpu memory ratio', default=None)
    parser.add_argument('--bseed', type=int, help='random seed for training backbone', default=0)
    parser.add_argument('--quick', action="store_true", help='ignore some evaluation')
    parser.add_argument('--no_tqdm', action="store_true", help='disable the tqdm')

    exp_args = parser.parse_args()
    use_tqdm = False if exp_args.no_tqdm else True
    # sub_metric = 'avg_loss'
    main(exp_args)



# https://github.com/pyg-team/pytorch_geometric/blob/master/examples/dgcnn_classification.py

from torch import Tensor
from typing import Callable, Union
from torch_geometric.typing import OptTensor, PairOptTensor, PairTensor

import torch
from torch.nn import Linear
import torch.nn.functional as F

from torch_geometric.nn.inits import reset
from torch_geometric.nn.conv import MessagePassing

from ..utils import FeatEncoder, MLP


class DGCNN(torch.nn.Module):
    def __init__(self, x_dim, pos_dim, model_config, feat_info, n_categorical_feat_to_use=-1, n_scalar_feat_to_use=-1, **kwargs):
        super().__init__()

        hidden_size = model_config['hidden_size']
        self.n_layers = model_config['n_layers']

        self.x_dim = x_dim
        self.pos_dim = pos_dim

        self.dropout_p = model_config['dropout_p']
        norm_type = model_config['norm_type']
        act_type = model_config['act_type']

        self.node_encoder = FeatEncoder(hidden_size, feat_info['node_categorical_feat'], feat_info['node_scalar_feat'], n_categorical_feat_to_use, n_scalar_feat_to_use)
        self.edge_encoder = FeatEncoder(hidden_size, feat_info['edge_categorical_feat'], feat_info['edge_scalar_feat'])

        self.convs = torch.nn.ModuleList()
        for _ in range(self.n_layers):
            mlp = MLP([hidden_size*3, hidden_size*2, hidden_size], 0.0, norm_type, act_type)
            self.convs.append(EdgeConv(mlp, hidden_size, norm_type, act_type, aggr='mean'))

    def forward(self, x, pos, edge_attr, edge_index, batch, edge_attn=None):
        if self.x_dim == 0 and self.pos_dim != 0:
            feats = pos
        elif self.x_dim != 0 and self.pos_dim == 0:
            feats = x
        elif self.x_dim == 0 and self.pos_dim == 0:
            feats = torch.ones(x.shape[0], 1, device=x.device)
        else:
            feats = torch.cat([x, pos], dim=1)

        x = self.node_encoder(feats)
        edge_attr = self.edge_encoder(edge_attr)
        for i in range(self.n_layers):
            identity = x
            x = self.convs[i](x, edge_index, batch=batch, edge_attr=edge_attr, edge_attn=edge_attn)
            x = x + identity
            x = F.dropout(x, self.dropout_p, training=self.training)
        return x


class EdgeConv(MessagePassing):
    r"""The dynamic edge convolutional operator from the `"Dynamic Graph CNN
    for Learning on Point Clouds" <https://arxiv.org/abs/1801.07829>`_ paper
    (see :class:`torch_geometric.nn.conv.EdgeConv`), where the graph is
    dynamically constructed using nearest neighbors in the feature space.

    Args:
        nn (torch.nn.Module): A neural network :math:`h_{\mathbf{\Theta}}` that
            maps pair-wise concatenated node features :obj:`x` of shape
            `:obj:`[-1, 2 * in_channels]` to shape :obj:`[-1, out_channels]`,
            *e.g.* defined by :class:`torch.nn.Sequential`.
        k (int): Number of nearest neighbors.
        aggr (string): The aggregation operator to use (:obj:`"add"`,
            :obj:`"mean"`, :obj:`"max"`). (default: :obj:`"max"`)
        num_workers (int): Number of workers to use for k-NN computation.
            Has no effect in case :obj:`batch` is not :obj:`None`, or the input
            lies on the GPU. (default: :obj:`1`)
        **kwargs (optional): Additional arguments of
            :class:`torch_geometric.nn.conv.MessagePassing`.

    Shapes:
        - **input:**
          node features :math:`(|\mathcal{V}|, F_{in})` or
          :math:`((|\mathcal{V}|, F_{in}), (|\mathcal{V}|, F_{in}))`
          if bipartite,
          batch vector :math:`(|\mathcal{V}|)` or
          :math:`((|\mathcal{V}|), (|\mathcal{V}|))`
          if bipartite *(optional)*
        - **output:** node features :math:`(|\mathcal{V}|, F_{out})` or
          :math:`(|\mathcal{V}_t|, F_{out})` if bipartite
    """
    def __init__(self, nn: Callable, hidden_size, norm_type, act_type, aggr: str = 'max', **kwargs):
        super().__init__(aggr=aggr, flow='source_to_target', **kwargs)

        self.nn = nn
        self.post_nn = Linear(hidden_size, hidden_size)
        self.act_fn = MLP.get_act(act_type)()
        self.norm = MLP.get_norm(norm_type)(hidden_size)
        self.reset_parameters()

    def reset_parameters(self):
        reset(self.nn)

    def forward(
            self, x: Union[Tensor, PairTensor], edge_index,
            batch, edge_attr=None, edge_attn=None) -> Tensor:
        ## type: (Tensor, OptTensor) -> Tensor  # noqa
        # type: (PairTensor, Optional[PairTensor]) -> Tensor  # noqa
        """"""
        if isinstance(x, Tensor):
            x: PairTensor = (x, x)

        if x[0].dim() != 2:
            raise ValueError("Static graphs not supported in DynamicEdgeConv")

        b: PairOptTensor = (None, None)
        if isinstance(batch, Tensor):
            b = (batch, batch)
        elif isinstance(batch, tuple):
            assert batch is not None
            b = (batch[0], batch[1])

        # propagate_type: (x: PairTensor)
        out = self.propagate(edge_index, x=x, size=None, edge_attr=edge_attr, edge_attn=edge_attn)
        out = self.post_nn(out)
        out = self.act_fn(self.norm(out))
        return out

    def message(self, x_i: Tensor, x_j: Tensor, edge_attr, edge_attn) -> Tensor:
        msg = self.nn(torch.cat([x_i, x_j - x_i, edge_attr], dim=-1))
        if edge_attn is not None:
            return msg * edge_attn
        else:
            return msg

    def __repr__(self) -> str:
        return f'{self.__class__.__name__}(nn={self.nn}, k={self.k})'



# https://pytorch-geometric.readthedocs.io/en/latest/_modules/torch_geometric/nn/models/gnn_explainer.html
# https://github.com/RexYing/gnn-model-explainer/blob/master/explainer/explain.py

import torch
import torch.nn as nn
from math import sqrt
from ..base import BaseRandom


class GNNExplainer(BaseRandom):

    def __init__(self, clf, criterion, config):
        super().__init__()
        self.name = 'gnnexplainer'
        self.clf = clf
        # self.extractor = extractor
        self.criterion = criterion
        self.device = next(self.parameters()).device

        self.pred_loss_coef = config['pred_loss_coef']
        self.size_loss_coef = config['size_loss_coef']
        self.mask_ent_loss_coef = config['mask_ent_loss_coef']
        self.iter_per_sample = config['iter_per_sample']
        self.iter_lr = config['iter_lr']

    def __loss__(self, mask, clf_logits, clf_labels, epoch):
        pred_loss = self.criterion(clf_logits, clf_labels.float())

        size_loss = mask.mean()
        mask_ent_reg = -mask * (mask + 1e-10).log() - (1 - mask) * (1 - mask + 1e-10).log()
        mask_ent_loss = mask_ent_reg.mean()

        pred_loss = self.pred_loss_coef * pred_loss
        size_loss = self.size_loss_coef * size_loss
        mask_ent_loss = self.mask_ent_loss_coef * mask_ent_loss

        loss = pred_loss + size_loss + mask_ent_loss
        loss_dict = {'loss': loss.item(), 'pred': pred_loss.item(), 'size': size_loss.item(), 'ent': mask_ent_loss.item()}
        return loss, loss_dict

    def _initialize_masks(self, x, init="normal"):
        N = x.size()[0]
        std = torch.nn.init.calculate_gain("relu") * sqrt(2.0 / (2 * N))
        self.node_mask = torch.nn.Parameter(torch.FloatTensor(N, 1).normal_(1, std))

    def _clear_masks(self):
        self.node_mask = None

    def forward_pass(self, data, epoch, **kwargs):

        self._clear_masks()
        try:
            _, edge_index = self.clf.get_emb(data)
        except: # for post-hoc explaining inherent models
            _, edge_index = self.clf.clf.get_emb(data)
        
        with torch.no_grad():
            original_clf_logits = self.clf(data)

        self._initialize_masks(data.x)
        self.to(data.x.device)

        parameters = [self.node_mask]
        optimizer = torch.optim.Adam(parameters, lr=self.iter_lr)

        for _ in range(self.iter_per_sample):
            optimizer.zero_grad()
            node_mask = torch.sigmoid(self.node_mask)
            edge_mask = self.node_attn_to_edge_attn(node_mask, edge_index)
            masked_clf_logits = self.clf(data, edge_attn=edge_mask)
            loss, loss_dict = self.__loss__(self.node_mask.sigmoid(), masked_clf_logits, original_clf_logits.sigmoid(), epoch)
            loss.backward()
            optimizer.step()

        return loss, loss_dict, original_clf_logits, self.node_mask.sigmoid().reshape(-1)

    @staticmethod
    def node_attn_to_edge_attn(node_attn, edge_index):
        src_attn = node_attn[edge_index[0]]
        dst_attn = node_attn[edge_index[1]]
        edge_attn = src_attn * dst_attn
        return edge_attn



import yaml
import json
import argparse
from tqdm import tqdm
from pathlib import Path
from sklearn.metrics import roc_auc_score
import scipy.stats
import torch
import torch.nn as nn
from torch.nn import functional as F
from eval import FidelEvaluation, AUCEvaluation
from get_model import Model
from baselines import PGExplainer, SubgraphX, PGMExplainer, GNNExplainer
from utils import to_cpu, log_epoch, get_data_loaders, set_seed, load_checkpoint, ExtractorMLP, get_optimizer
from utils import inherent_models, post_hoc_explainers, post_hoc_attribution, name_mapping
import torchmetrics
from statistics import mean
import numpy as np
import pandas as pd
import warnings
from datetime import datetime
warnings.filterwarnings("ignore")


def eval_one_batch(baseline, data, epoch):
    with torch.set_grad_enabled(baseline.name in ['gradcam', 'gradgeo', 'gnnexplainer']):
        baseline.extractor.eval() if hasattr(baseline, 'extractor') else None
        baseline.clf.eval()

        # do_sampling = True if phase == 'valid' and baseline.name == 'pgexplainer' else False # we find this is better for BernMaskP
        loss, loss_dict, infer_clf_logits, node_attn = baseline.forward_pass(data, epoch=epoch, do_sampling=False)
        # from sklearn.metrics import roc_auc_score
        # roc_auc_score(infer_clf_logits, data.y.cpu())
        return loss_dict, to_cpu(infer_clf_logits), to_cpu(node_attn)


def one_seed_attn(baseline, data_loader, epoch, phase, seed, signal_class, writer=None, metric_list=None):
    # log_epoch(seed, epoch, phase, avg_loss_dict, epoch_dict, writer)
    return

def test_uncertainty(config, method_name, model_name, backbone_seed, method_seeds, dataset_name, log_dir, device):
    set_seed(backbone_seed, method_name=method_name, backbone_name=model_name)
    writer = None
    print('The logging directory is', log_dir), print('=' * 80)

    # batch_size = config['optimizer']['batch_size']
    batch_size = 1
    data_config = config['data']
    loaders, test_set, dataset = get_data_loaders(dataset_name, batch_size, data_config, dataset_seed=0)
    signal_class = dataset.signal_class

    clf = Model(model_name, config[model_name],  # backbone_config
                method_name, config[method_name],  # method_config
                dataset).to(device)

    extractor = ExtractorMLP(config[model_name]['hidden_size'], config[method_name], config['data'].get('use_lig_info', False)) \
        if method_name in inherent_models + ['pgexplainer'] else nn.Identity()
    extractor = extractor.to(device)
    criterion = F.binary_cross_entropy_with_logits

    # establish the model and the metrics

    metric_list = [AUCEvaluation()] if batch_size > 64 else None

    pbar, avg_loss_dict = tqdm(loaders['test']), dict()
    [eval_metric.reset() for eval_metric in metric_list] if metric_list else None

    clf_ACC, clf_AUC = torchmetrics.Accuracy(task='binary'), torchmetrics.AUROC(task='binary')
    map_location = torch.device('cpu') if not torch.cuda.is_available() else None
    epoch_avg_attn, epoch_std_attn, epoch_label, sig_spear, bkg_spear, all_spear = [], [], [], [], [], []
    # sp_list = []
    multi_epoch_attn = None
    for idx, data in enumerate(pbar):
        if data.y.item() == 0:
            continue

        multi_attn = None
        eval_dict = {}
        for method_seed in method_seeds:
            baseline = eval(name_mapping[method_name])(clf, criterion, config[method_name]) \
                if method_name != 'pgexplainer' else PGExplainer(clf, extractor, criterion, config['pgexplainer'])
            if method_name == 'pgexplainer':
                assert load_checkpoint(baseline, log_dir, model_name=method_name, seed=method_seed, map_location=map_location,
                                   backbone_seed=backbone_seed, verbose=False)
            else:
                pass
                # print(f'Running {method_name} for seed {method_seed}.')

            erm_dir = log_dir.parent / 'erm'
            assert load_checkpoint(baseline.clf, erm_dir, model_name='erm', seed=backbone_seed, map_location=map_location, verbose=False)
            # data = negative_augmentation(data, data_config, phase, data_loader, idx, loader_len)
            loss_dict, clf_logits, attn = eval_one_batch(baseline, data.to(baseline.device), 1)
            ex_labels, clf_labels, data = to_cpu(data.node_label), to_cpu(data.y), data.cpu()
            # print()
            eval_dict.update({'clf_acc': clf_ACC(clf_logits, clf_labels), 'clf_auc': clf_AUC(clf_logits, clf_labels)})
            eval_dict.update({f'exp_auc_{method_seed}': roc_auc_score(ex_labels, attn)})
            multi_attn = attn.unsqueeze(1) if multi_attn is None else torch.cat([multi_attn, attn.unsqueeze(1)], dim=1)

        info_graph = torch.full((multi_attn.shape[0], 1), idx)
        multi_attn_plus = torch.cat([multi_attn, ex_labels.unsqueeze(1), info_graph], dim=1)
        multi_epoch_attn = multi_attn_plus if multi_epoch_attn is None else torch.cat([multi_epoch_attn, multi_attn_plus], dim=0)

        avg_attn, std_attn = torch.mean(multi_attn, dim=1), torch.std(multi_attn, dim=1)
        batch_exp_auc_avg, batch_exp_auc_std = roc_auc_score(ex_labels, avg_attn), roc_auc_score(ex_labels, 1-std_attn)
        batch_spearman = scipy.stats.spearmanr(avg_attn, std_attn)[0]
        sig_index, bkg_index = torch.where(data.node_label == 1), torch.where(data.node_label == 0)
        batch_sig_spear = scipy.stats.spearmanr(avg_attn[sig_index], std_attn[sig_index])[0]
        batch_bkg_spear = scipy.stats.spearmanr(avg_attn[bkg_index], std_attn[bkg_index])[0]
        eval_dict.update({'exp_auc_avg': batch_exp_auc_avg, 'exp_auc_std': batch_exp_auc_std, 'spearman': batch_spearman})

        epoch_avg_attn.append(avg_attn), epoch_std_attn.append(std_attn), epoch_label.append(ex_labels)
        all_spear.append(batch_spearman), sig_spear.append(batch_sig_spear), bkg_spear.append(batch_bkg_spear)

        # print('The spearman correlation between avg and std:', )
        desc = log_epoch(backbone_seed, 1, 'test', loss_dict, eval_dict, phase_info='dataset')
        pbar.set_description(desc)
        # compute the avg_loss for the epoch desc
        exec('for k, v in loss_dict.items():\n\tavg_loss_dict[k]=(avg_loss_dict.get(k, 0) * idx + v) / (idx + 1)')

    # epoch_dict = {eval_metric.name: eval_metric.eval_epoch(return_att=True) for eval_metric in metric_list} if metric_list else {}
    # assert or  metric_list[0].name == 'exp_auc'
    epoch_avg_attn, epoch_std_attn, epoch_label = torch.cat(epoch_avg_attn), torch.cat(epoch_std_attn), torch.cat(epoch_label)
    if method_name in ['subgraphx', 'pgexplainer']:
        save_attn = pd.DataFrame(multi_epoch_attn, columns=method_seeds + ['node_labels', 'graph_id'])
        # indexes =
        save_dir = log_dir / f"bseed_{backbone_seed}_{method_name}_attns.csv"
        save_attn.to_csv(save_dir)

    epoch_dict = {'exp_auc_avg': roc_auc_score(epoch_label, epoch_avg_attn),
                  'exp_auc_std': roc_auc_score(epoch_label, 1-epoch_std_attn),
                  'all_spearman': sum(all_spear) / len(all_spear),
                  'sig_spearman': sum(sig_spear) / len(sig_spear),
                  'bkg_spearman': sum(bkg_spear) / len(bkg_spear)}
                  # 'all_spearman': scipy.stats.spearmanr(epoch_avg_attn, epoch_std_attn)[0]
    epoch_dict.update({'clf_acc': clf_ACC.compute().item(), 'clf_auc': clf_AUC.compute().item()})
    print('-' * 80, '\n', 'epoch_dict:', epoch_dict, '\n', '-' * 80)

    return epoch_dict

def save_multi_bseeds(multi_methods_res, method_name, exp_method, seeds):
    seeds += ['avg', 'std']
    indexes = [method_name+'_'+str(seed) for seed in seeds]
    # from itertools import product
    # indexes = ['_'.join(item) for item in product(methods, seeds)]
    df = pd.DataFrame(multi_methods_res, index=indexes)

    day_dir = Path('../result') / config_name / datetime.now().strftime("%m_%d")
    day_dir.mkdir(parents=True, exist_ok=True)
    csv_dir = day_dir / ('_'.join([exp_method, method_name, 'sensitivity.csv']))
    with open(csv_dir, mode='w') as f:
        df.to_csv(f, lineterminator="\n", header=f.tell() == 0)
    return df


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='EXP')
    parser.add_argument('-d', '--dataset', type=str, help='dataset used', default='actstrack')
    parser.add_argument('--method', type=str, help='method used')
    parser.add_argument('-b', '--backbone', type=str, help='backbone used', default='egnn')
    parser.add_argument('--cuda', type=int, help='cuda device id, -1 for cpu', default=-1)
    parser.add_argument('--note', type=str, help='note in log name', default='')
    parser.add_argument('--mseeds', type=int, nargs="+", help='random seed for training explainer', default=[0, 1, 2, 3, 4, 5, 6, 7, 8, 9])
    parser.add_argument('--bseed', type=int, nargs="+", help='random seed for training backbone', default=0)
    # parser.add_argument('--mseed', type=int, help='random seed for explainer', default=0)

    args = parser.parse_args()
    # sub_metric = 'avg_loss'
    print(args)

    dataset_name, method_name, model_name, cuda_id, note = args.dataset, args.method, args.backbone, args.cuda, args.note
    config_name = '_'.join([model_name, dataset_name])
    # sub_dataset_name = '_' + dataset_name.split('_')[1] if len(dataset_name.split('_')) > 1 else ''
    config_path = Path('../configs') / f'{config_name}.yml'
    config = yaml.safe_load((config_path).open('r'))
    if config[method_name].get(model_name, False):
        config[method_name].update(config[method_name][model_name])

    print('=' * 80)
    print(f'Config for {method_name}: ', json.dumps(config[method_name], indent=4))

    # the directory is like egnn_actstrack / bseed0 / lri_bern
    device = torch.device(f'cuda:{cuda_id}' if cuda_id >= 0 and torch.cuda.is_available() else 'cpu')
    all_seeds_res = []
    model_dir = Path('../log') / config_name / method_name
    # cuda_id = args.cuda
    # for method_seed in args.mseeds:
    for backbone_seed in args.bseed:
        epoch_dict = test_uncertainty(config, method_name, model_name, backbone_seed, args.mseeds, dataset_name, model_dir, device)
    all_seeds_res += [epoch_dict]
    df = pd.DataFrame(all_seeds_res, index=['_'.join([f'bseed{bseed}', method_name]) for bseed in args.bseed])
    with open(f'result/egnn_synmol/uncertainty/{method_name}.csv', mode='a') as f:
        df.to_csv(f, lineterminator="\n", header=f.tell() == 0)

    # multi_seeds_res += [avg_report, std_report]
    # print(json.dumps(avg_std_report, indent=4))
    # save_multi_bseeds(multi_seeds_res, method_name, exp_method, args.bseeds)




# https://github.com/pyg-team/pytorch_geometric/blob/72eb1b38f60124d4d700060a56f7aa9a4adb7bb0/torch_geometric/nn/models/pg_explainer.py

import torch
import torch.nn as nn
from torch_scatter import scatter_mean
from ..base import BaseRandom


class PGExplainer(BaseRandom):

    def __init__(self, clf, extractor, criterion, config):
        super().__init__()
        self.name = 'pgexplainer'
        self.clf = clf
        self.extractor = extractor
        self.criterion = criterion
        self.device = next(self.parameters()).device

        self.pred_loss_coef = config['pred_loss_coef']
        self.size_loss_coef = config['size_loss_coef']
        self.mask_ent_loss_coef = config['mask_ent_loss_coef']
        self.temp_tuple = config['temp']
        self.epochs = config['epochs']

    def __loss__(self, mask, clf_logits, clf_labels, epoch, batch):
        pred_loss = self.criterion(clf_logits, clf_labels.float())

        # size_loss = attn.sum()
        size_loss = mask.mean()  # take mean makes it easier to tune coef for different datasets
        mask_ent_reg = -mask * (mask + 1e-10).log() - (1 - mask) * (1 - mask + 1e-10).log()
        # mask_ent_loss = mask_ent_reg.mean() if batch is None else scatter_mean(mask_ent_reg, batch.reshape(-1, 1)).sum()
        mask_ent_loss = mask_ent_reg.mean() if batch is None else scatter_mean(mask_ent_reg, batch.reshape(-1, 1)).mean()  # take mean makes it easier to tune coef for different datasets

        pred_loss = self.pred_loss_coef * pred_loss
        size_loss = self.size_loss_coef * size_loss
        mask_ent_loss = self.mask_ent_loss_coef * mask_ent_loss

        loss = pred_loss + size_loss + mask_ent_loss
        loss_dict = {'loss': loss.item(), 'pred': pred_loss.item(), 'size': size_loss.item(), 'ent': mask_ent_loss.item()}
        return loss, loss_dict

    def forward_pass(self, data, epoch, do_sampling):
        try:
            emb, edge_index = self.clf.get_emb(data)
        except: # for post-hoc explaining inherent models
            emb, edge_index = self.clf.clf.get_emb(data) 
        node_mask_log_logits = self.extractor(emb, batch=data.batch, pool_out_lig=None)

        node_mask = self.sampling(node_mask_log_logits, do_sampling, epoch)
        edge_mask = self.node_attn_to_edge_attn(node_mask, edge_index)

        original_clf_logits = self.clf(data)
        masked_clf_logits = self.clf(data, edge_attn=edge_mask)

        loss, loss_dict = self.__loss__(node_mask_log_logits.sigmoid(), masked_clf_logits, original_clf_logits.sigmoid(), epoch, data.batch)
        return loss, loss_dict, original_clf_logits, node_mask.reshape(-1)

    def get_temp(self, e: int) -> float:
        temp = self.temp_tuple
        return temp[0] * ((temp[1] / temp[0])**(e / self.epochs))

    def sampling(self, attn_log_logits, do_sampling, epoch):
        if do_sampling:
            random_noise = torch.empty_like(attn_log_logits).uniform_(1e-10, 1 - 1e-10)
            random_noise = torch.log(random_noise) - torch.log(1.0 - random_noise)

            temp = self.get_temp(epoch)
            attn_bern = ((attn_log_logits + random_noise) / temp).sigmoid()
        else:
            attn_bern = (attn_log_logits).sigmoid()
        return attn_bern

    @staticmethod
    def node_attn_to_edge_attn(node_attn, edge_index):
        src_attn = node_attn[edge_index[0]]
        dst_attn = node_attn[edge_index[1]]
        edge_attn = src_attn * dst_attn
        return edge_attn



# https://github.com/pyg-team/pytorch_geometric/blob/72eb1b38f60124d4d700060a56f7aa9a4adb7bb0/torch_geometric/nn/models/pg_explainer.py

import torch
import torch.nn as nn
import torch.nn.functional as F
import random
from torch_geometric.data import Data, Batch
class Orphicx(nn.Module):

    def __init__(self, clf, extractor, criterion, config):
        super().__init__()
        self.name = 'orphicx'
        self.clf = clf
        self.extractor = extractor
        self.criterion = criterion
        self.device = next(self.parameters()).device
        self.sample_num = {'Nalpha': 25, 'Nbeta' : 25, 'NX': 2, 'NA': 1}
        self.causal_dim = 8
        self.spurious_dim = extractor.encode_size - self.causal_dim

        # self.pred_loss_coef = config.pred_loss_coef
        self.size_loss_coef = config.size_loss_coef
        self.kl_loss_coef = config.kl_loss_coef
        self.vgae_loss_coef = config.vgae_loss_coef

    def __loss__(self, att_adj, recovered_adj, org_adj, att_logits, org_logits, mu, logvar, data):

        size_loss = att_adj.mean()  # take mean makes it easier to tune coef for different datasets
        kl_loss = F.kl_div(F.log_softmax(att_logits, dim=1), F.softmax(org_logits, dim=1), reduction='batchmean')
            # self.criterion(recovered_adj.reshape(-1), org_adj.reshape(-1))
        vgae_loss = self.extractor.vgae_loss(recovered_adj, org_adj, mu, logvar, data)


        size_loss = self.size_loss_coef * size_loss
        kl_loss = self.kl_loss_coef * kl_loss
        vgae_loss = self.vgae_loss_coef * vgae_loss

        loss =  size_loss + kl_loss + vgae_loss
        loss_dict = {'loss': loss.item(), 'size': size_loss.item(), 'kl': kl_loss.item(), 'vgae':vgae_loss.item()}
        return loss, loss_dict

    def warming(self, data):
        clf_logits = self.clf(data)
        pred_loss = self.criterion(clf_logits, data.y.float())
        return pred_loss, clf_logits

    def explaining(self, data, x_level, do_sampling):
        original_clf_logits = self.clf(data, edge_attr=data.edge_attr)
        mu, logvar = self.extractor.encoder(data.x, data.edge_index) ####
        all_z = self.extractor.encode(data.x, data.edge_index)

        caul_z = torch.zeros_like(all_z)
        caul_z[:,:self.causal_dim] = all_z[:,:self.causal_dim]

        org_adj = self.get_adj(data)
        attn_adj = self.extractor.decode(caul_z)
        recovered_adj = self.extractor.decode(all_z)
        edge_attn = torch.zeros([data.num_edges, 1])

        for index, (i, j) in enumerate(data.edge_index.T):
            edge_attn[index, 0] = attn_adj[i, j]

        if x_level == 'edge':
            att = edge_attn
        else:
            att = self.edge_attn_to_node_attn(edge_attn, data)
        # edge_attn = self.node_attn_to_edge_attn(att, data.edge_index) if att_type == 'node' else att if att_type == 'edge' else None

        masked_clf_logits = self.clf(data, edge_attr=data.edge_attr, edge_attn=edge_attn)

        # batch = data.batch if att_type == 'node' else data.batch[data.edge_index[0]]

        other_loss, loss_dict = self.__loss__(attn_adj, recovered_adj, org_adj, masked_clf_logits, original_clf_logits, mu, logvar, data)
        loss_dict['caul_loss'] = self.compute_information_flow(data, self.extractor.decoder, self.clf).item()
        loss = other_loss + loss_dict['caul_loss']

        # att = node_mask.reshape(-1) if x_level == 'node' else edge_mask.reshape(-1)
        return loss, loss_dict, original_clf_logits, att.reshape(-1)

    def compute_information_flow(self, data, decoder, clf):
        # batch_loss = 0
        # for graph in data.to_data_list():
        causal_loss = []
        sub_adj_list = [self.get_adj(graph, sub_loop=True) for graph in data.to_data_list()]
        feat_list = [graph.x for graph in data.to_data_list()]
        pos_list = [graph.pos for graph in data.to_data_list()]

        for idx in random.sample(range(len(feat_list)), self.sample_num['NX']):
            causal_loss += [self.joint_uncond(decoder, clf, sub_adj_list[idx], feat_list[idx], pos_list[idx], act=torch.sigmoid)]
            # causal_loss += _causal_loss
            for A_idx in random.sample(range(len(feat_list)), self.sample_num['NA']-1):
                causal_loss += [self.joint_uncond(decoder, clf, sub_adj_list[A_idx], feat_list[idx], pos_list[idx], act=torch.sigmoid)]

        return torch.stack(causal_loss).mean()
        # batch_loss += graph_loss
    """
    joint_uncond:
        Sample-based estimate of "joint, unconditional" causal effect, -I(alpha; Yhat).
    Inputs:
        - params['Nalpha'] monte-carlo samples per causal factor
        - params['Nbeta']  monte-carlo samples per noncausal factor
        - params['K']      number of causal factors
        - params['L']      number of noncausal factors
        - params['M']      number of classes (dimensionality of classifier output)
        - decoder
        - classifier
        - device
    Outputs:
        - negCausalEffect (sample-based estimate of -I(alpha; Yhat))
        - info['xhat']
        - info['yhat']
    """


    def joint_uncond(self, decoder, classifier, adj, feat, pos, act=torch.sigmoid, mu=0, std=1,
                     device=None):
        params = self.sample_num
        params['M'] = self.extractor.output_size

        eps = 1e-8
        I = 0.0
        q = torch.zeros(params['M'], device=device)
        feat = feat.repeat(params['Nalpha'] * params['Nbeta'], 1, 1)
        adj = adj.repeat(params['Nalpha'] * params['Nbeta'], 1, 1)
        all_pos = pos.repeat(params['Nalpha'] * params['Nbeta'], 1, 1)
        if torch.is_tensor(mu):
            alpha_mu = mu[:, :self.causal_dim]
            beta_mu = mu[:, self.causal_dim:]

            alpha_std = std[:, :self.causal_dim]
            beta_std = std[:, self.causal_dim:]
        else:
            alpha_mu = 0
            beta_mu = 0
            alpha_std = 1
            beta_std = 1

        alpha = torch.randn((params['Nalpha'], adj.shape[-1], self.causal_dim), device=device).mul(alpha_std).add_(
            alpha_mu).repeat(1, params['Nbeta'], 1).view(params['Nalpha'] * params['Nbeta'], adj.shape[-1], self.causal_dim)
        beta = torch.randn((params['Nalpha'] * params['Nbeta'], adj.shape[-1], self.spurious_dim), device=device).mul(
            beta_std).add_(beta_mu)
        zs = torch.cat([alpha, beta], dim=-1)
        xhat = act(decoder(zs)) * adj
        data_list = []
        for index in range(zs.shape[0]):
            x = zs[index]
            one_pos = all_pos[index]
            edge_index = self.get_edge_index(xhat[index])
            edge_attn = self.get_edge_attn(xhat[index], edge_index)
            data_list += [Data(x=x, edge_index=edge_index, pos=one_pos, edge_attn=edge_attn)]
        data = Batch.from_data_list(data_list)
        # if node_idx is None:
        logits = classifier(data, edge_attn=data.edge_attn, skip_emb=True)
        # else:
        # logits = torch.cat()
        #     logits = classifier(feat, self.get_edge_index(xhat))[:, node_idx, :]
        # two-classes
        yhat = torch.cat([torch.sigmoid(logits) , 1-torch.sigmoid(logits)], dim=1)
        yhat = yhat.view(params['Nalpha'], params['Nbeta'], 2)

        p = yhat.mean(1)
        I = torch.sum(torch.mul(p, torch.log(p + eps)), dim=1).mean()
        q = p.mean(0)
        I = I - torch.sum(torch.mul(q, torch.log(q + eps)))
        return -I

    @staticmethod
    def edge_attn_to_node_attn(edge_attn, data):
        edge_index, num_nodes = data.edge_index, data.num_nodes
        node_attn = torch.zeros([num_nodes, 1])
        for index, (i, j) in enumerate(edge_index.T):
            node_attn[i] += edge_attn[index, 0]
            node_attn[j] += edge_attn[index, 0]
        # edge_attn = 1 - (1-src_attn) * (1-dst_attn)
        max_ = max(node_attn)
        min_ = min(node_attn)
        node_attn = (node_attn - min_) / (max_ - min_)
        return node_attn

    @staticmethod
    def get_adj(data, sub_loop=False):
        if sub_loop:
            adj = torch.zeros([data.num_nodes, data.num_nodes])
        else:
            adj = torch.eye(data.num_nodes)
        for i, j in data.edge_index.T:
            adj[i, j] = 1
            adj[j, i] = 1
        return adj

    @staticmethod
    def get_edge_index(matrix):
        """输入图的邻接矩阵，输出邻接表"""
        col, row = [], []
        N = len(matrix)
        for i in range(N):
            tmp1 = {}
            for j in range(i):
                if (matrix[i][j] and (i != j)):
                    col += [i]
                    row += [j]
        col = torch.tensor(col).unsqueeze(0)
        row = torch.tensor(row).unsqueeze(0)
        return torch.cat([col, row], dim=0)

    @staticmethod
    def get_edge_attn(matrix, edge_index):
        edge_attr = torch.zeros([edge_index.shape[1], 1])
        for index, edge_pair in enumerate(edge_index.T):
            i, j = edge_pair
            edge_attr[index, 0] = matrix[i][j]

        return edge_attr


import os
import sys
import random
from tqdm import tqdm
from joblib import Parallel, delayed
import torch
import numpy as np
from Bio.PDB import ShrakeRupley
from statistics import mean

inherent_models = ['asap', 'lri_bern', 'lri_gaussian', 'vgib', 'ciga']
post_hoc_explainers = ['pgexplainer', 'gnnexplainer', 'subgraphx', 'pgmexplainer']
post_hoc_attribution = ['gradcam', 'gnnlrp', 'gradx', 'inter_grad']

name_mapping = {'lri_bern': 'LRIBern', 'lri_gaussian': 'LRIGaussian', 'vgib': 'VGIB', 'ciga': 'CIGA', 'dir': 'DIR',
                'gradcam': 'GradCAM', 'inter_grad': 'InterGrad', 'gradx': 'GradX', 'gnnlrp': 'GNNLRP',
                'pgexplainer': 'PGExplainer', 'gnnexplainer': 'GNNExplainer', 'subgraphx': 'SubgraphX',
                'pgmexplainer': 'PGMExplainer',
                'test': 'Test', 'asap': 'ASAP'}


sr = ShrakeRupley(probe_radius=1.4,  # in A. Default is 1.40 roughly the radius of a water molecule.
                  n_points=100)  # resolution of the surface of each atom. Default is 100. A higher number of points results in more precise measurements, but slows down the calculation.


allowable_features = {
    'possible_atomic_num_list': list(range(1, 119)) + ['misc'],
    'possible_chirality_list': [
        'CHI_UNSPECIFIED',
        'CHI_TETRAHEDRAL_CW',
        'CHI_TETRAHEDRAL_CCW',
        'CHI_OTHER'
    ],
    'possible_degree_list': [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 'misc'],
    'possible_numring_list': [0, 1, 2, 3, 4, 5, 6, 'misc'],
    'possible_implicit_valence_list': [0, 1, 2, 3, 4, 5, 6, 'misc'],
    'possible_formal_charge_list': [-5, -4, -3, -2, -1, 0, 1, 2, 3, 4, 5, 'misc'],
    'possible_numH_list': [0, 1, 2, 3, 4, 5, 6, 7, 8, 'misc'],
    'possible_number_radical_e_list': [0, 1, 2, 3, 4, 'misc'],
    'possible_hybridization_list': [
        'SP', 'SP2', 'SP3', 'SP3D', 'SP3D2', 'misc'
    ],
    'possible_is_aromatic_list': [False, True],
    'possible_is_in_ring3_list': [False, True],
    'possible_is_in_ring4_list': [False, True],
    'possible_is_in_ring5_list': [False, True],
    'possible_is_in_ring6_list': [False, True],
    'possible_is_in_ring7_list': [False, True],
    'possible_is_in_ring8_list': [False, True],
    'possible_amino_acids': ['ALA', 'ARG', 'ASN', 'ASP', 'CYS', 'GLN', 'GLU', 'GLY', 'HIS', 'ILE', 'LEU', 'LYS', 'MET',
                             'PHE', 'PRO', 'SER', 'THR', 'TRP', 'TYR', 'VAL', 'HIP', 'HIE', 'TPO', 'HID', 'LEV', 'MEU',
                             'PTR', 'GLV', 'CYT', 'SEP', 'HIZ', 'CYM', 'GLM', 'ASQ', 'TYS', 'CYX', 'GLZ', 'misc'],
    'possible_atom_type_2': ['C*', 'CA', 'CB', 'CD', 'CE', 'CG', 'CH', 'CZ', 'N*', 'ND', 'NE', 'NH', 'NZ', 'O*', 'OD',
                             'OE', 'OG', 'OH', 'OX', 'S*', 'SD', 'SG', 'misc'],
    'possible_atom_type_3': ['C', 'CA', 'CB', 'CD', 'CD1', 'CD2', 'CE', 'CE1', 'CE2', 'CE3', 'CG', 'CG1', 'CG2', 'CH2',
                             'CZ', 'CZ2', 'CZ3', 'N', 'ND1', 'ND2', 'NE', 'NE1', 'NE2', 'NH1', 'NH2', 'NZ', 'O', 'OD1',
                             'OD2', 'OE1', 'OE2', 'OG', 'OG1', 'OH', 'OXT', 'SD', 'SG', 'misc'],
}

def sum_fid_score(all_dict):
    sub_dict = {k: v for k, v in all_dict.items() if '@' in k}
    fidel_type = set([i.split('@')[0] for i in sub_dict.keys()])
    for prefix in fidel_type:
        fid_list = [all_dict[k] for k in all_dict if prefix in k]
        all_dict[prefix] = mean(fid_list)
    for prefix in fidel_type:
        match_fix = prefix.replace('+', '-')
        if '+' in prefix and match_fix in fidel_type:
            name = prefix.replace('+', '')
            all_dict[name] = all_dict[prefix] - all_dict[match_fix]


def set_seed(seed, method_name=None, backbone_name=None):
    # from pytorch_lightning import seed_everything
    # # set the random seeds.
    # seed_everything(seed)
    # torch.backends.cudnn.determinstic = True
    # torch.backends.cudnn.benchmark = False

    os.environ['PYTHONHASHSEED'] = str(seed)
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    torch.backends.cudnn.benchmark = False
    torch.backends.cudnn.deterministic = True
    os.environ['CUBLAS_WORKSPACE_CONFIG'] = ':4096:8'

    if method_name not in ['asap'] and backbone_name not in ['pointtrans']:
        torch.use_deterministic_algorithms(True)
    else:
        
        torch.use_deterministic_algorithms(False, warn_only=True)

def to_cpu(tensor):
    return tensor.detach().cpu() if tensor is not None else None


def safe_index(l, e):
    try:
        return l.index(e)
    except:
        return len(l) - 1


class HiddenPrints:
    def __enter__(self):
        self._original_stdout = sys.stdout
        sys.stdout = open(os.devnull, 'w')

    def __exit__(self, exc_type, exc_val, exc_tb):
        sys.stdout.close()
        sys.stdout = self._original_stdout


def disable_rdkit_logging():
    """
    Disables RDKit whiny logging.
    """
    import rdkit.rdBase as rkrb
    import rdkit.RDLogger as rkl
    logger = rkl.logger()
    logger.setLevel(rkl.ERROR)
    rkrb.DisableLog('rdApp.error')


def pmap_multi(pickleable_fn, data, n_jobs, verbose=1, desc=None, **kwargs):
  results = Parallel(n_jobs=n_jobs, verbose=verbose, timeout=None, )(
    delayed(pickleable_fn)(*d, **kwargs) for i, d in tqdm(enumerate(data), desc=desc)
  )

  return results


def get_random_idx_split(dataset_len, split, seed):
    np.random.seed(seed)

    print('[INFO] Randomly split dataset!')
    idx = np.arange(dataset_len)
    np.random.shuffle(idx)

    n_train, n_valid = int(split['train'] * len(idx)), int(split['valid'] * len(idx))
    train_idx = idx[:n_train]
    valid_idx = idx[n_train:n_train+n_valid]
    test_idx = idx[n_train+n_valid:]
    return {'train': train_idx, 'valid': valid_idx, 'test': test_idx}
